# cython: profile=True
cimport cython
from cython.parallel cimport prange
from libc.math cimport floor, sqrt
from libc.stdlib cimport malloc, free
from libc.stdio cimport printf
from cpython cimport PyErr_SetString

from cpython.object cimport PyObject
from cpython.sequence cimport PySequence_Fast, PySequence_Fast_ITEMS
from cpython.list cimport PyList_Size, PyList_GET_ITEM, PyList_Sort
from cpython.tuple cimport PyTuple_GET_ITEM, PyTuple_Size

from ms_peak_picker._c.peak_index cimport PeakIndex

from ms_peak_picker._c.peak_set cimport PeakBase, FittedPeak, PeakSet
from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet, DeconvolutedPeakSetIndexed

cimport numpy as cnp
import numpy as np

cdef object zeros = np.zeros


ctypedef fused peak_collection:
    PeakSet
    PeakIndex
    DeconvolutedPeakSet
    object


cpdef enum SimilarityMetrics:
    dot_product
    normalized_dot_product


cpdef enum PeakType:
    Fitted
    Deconvoluted


@cython.freelist(10000)
@cython.final
cdef class PeakMatch(object):
    cdef:
        public PeakBase peak_a
        public PeakBase peak_b
        public double score

    def __init__(self, peak_a, peak_b, score=0.0):
        self.peak_a = peak_a
        self.peak_b = peak_b
        self.score = score

    cpdef bint _lt(self, PeakMatch other):
        return self.score < other.score

    cpdef bint _gt(self, PeakMatch other):
        return self.score > other.score

    cpdef bint _eq(self, PeakMatch other):
        return self.peak_a == other.peak_a and self.peak_b == other.peak_b

    def __richcmp__(self, PeakMatch other, int code):
        if other is None:
            if code == 3:
                return True
            else:
                return False

        if code == 0:
            return self._lt(other)
        elif code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))
        elif code == 4:
            return self._gt(other)

    @staticmethod
    cdef PeakMatch _create(PeakBase peak_a, PeakBase peak_b, double score=0.0):
        cdef PeakMatch self = PeakMatch.__new__(PeakMatch)
        self.peak_a = peak_a
        self.peak_b = peak_b
        self.score = score
        return self

    def __getitem__(self, i):
        if i == 0:
            return self.peak_a
        elif i == 1:
            return self.peak_b
        else:
            raise IndexError(i)

    def __len__(self):
        return 2

    def __repr__(self):
        template = "{self.__class__.__name__}({self.peak_a}, {self.peak_b}, {self.score})"
        return template.format(self=self)


cdef PyObject** _make_fast_array(peak_collection peak_set):
    cdef:
        PyObject** items

    if peak_collection is PeakIndex:
        items = PySequence_Fast_ITEMS(
            PySequence_Fast(peak_set.peaks.peaks, "Error, could not create fast array from PeakIndex"))
    elif peak_collection is PeakSet:
        items = PySequence_Fast_ITEMS(
            PySequence_Fast(peak_set.peaks, "Error, could not create fast array from PeakSet"))
    elif peak_collection is DeconvolutedPeakSet:
        items = PySequence_Fast_ITEMS(
            PySequence_Fast(peak_set.peaks, "Error, could not create fast array from DeconvolutedPeakSet"))
    else:
        # This block is unsafe for any type that is not a list or a tuple, risking segfaults.
        # The list/tuple must have a lifespan exceeding the duration of this function call.
        items = PySequence_Fast_ITEMS(
            PySequence_Fast(peak_set, "Error, could not create fast array from object"))
    return items


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:
    """Computes the normalized dot product, also called cosine similarity between
    two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).

    Parameters
    ----------
    peak_set_a : Iterable of Peak-like
    peak_set_b : Iterable of Peak-like
        The two peak collections to compare. It is usually only useful to
        compare the similarity of peaks of the same class, so the types
        of the elements of `peak_set_a` and `peak_set_b` should match.
    precision : int, optional
        The precision of rounding to use when binning spectra. Defaults to 0

    Returns
    -------
    float
        The similarity between peak_set_a and peak_set_b. Between 0.0 and 1.0
    """

    cdef:
        tuple binned_spectra
        list _hold_a, _hold_b
        double n_a, n_b, n_ab, z, a, b
        double hi, scaler
        PeakBase peak
        double *bin_a
        double *bin_b
        long i, n, k, index, n_peaks_a, n_peaks_b
        PyObject** items_a
        PyObject** items_b
        PyObject* p_peak

    n_a = 0
    n_b = 0
    a = 0
    b = 0
    z = 0

    # Determine maximum value
    hi = 0
    if peak_collection is PeakIndex:
        peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is object:
        hi = 0
        for obj in peak_set_a:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
        for obj in peak_set_b:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
    elif peak_collection is PeakSet:
        peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
        peak = peak_set_a._mz_ordered[-1]
        hi = peak.mz
        peak = peak_set_b._mz_ordered[-1]
        hi = max(peak.mz, hi)

    # Calculate bin array size and allocate memory
    scaler = 10 ** precision
    k = <long>((hi * scaler) + 1)
    bin_a = <double*>malloc(k * sizeof(double))
    bin_b = <double*>malloc(k * sizeof(double))
    if bin_a == NULL or bin_b == NULL:
        if bin_a != NULL:
            free(bin_a)
        if bin_b != NULL:
            free(bin_b)
        PyErr_SetString(MemoryError, "Could not allocate bins")
        return -1
    for i in range(k):
        bin_a[i] = bin_b[i] = 0

    # Fill bins
    if peak_collection is object:
        _hold_a = list(peak_set_a)
        n_peaks_a = len(_hold_a)
    else:
        n_peaks_a = peak_set_a.get_size()

    if peak_collection is object:
        _hold_b = list(peak_set_b)
        n_peaks_b = len(_hold_b)
    else:
        n_peaks_b = peak_set_b.get_size()

    if peak_collection is object:
        items_a = _make_fast_array(_hold_a)
        items_b = _make_fast_array(_hold_b)
    else:
        items_a = _make_fast_array(peak_set_a)
        items_b = _make_fast_array(peak_set_b)
    with nogil:
        # Fill bins
        for i in range(n_peaks_a):
            p_peak = items_a[i]
            index = int((<PeakBase>p_peak).mz * scaler)
            bin_a[index] += (<PeakBase>p_peak).intensity

        for i in range(n_peaks_b):
            p_peak = items_b[i]
            index = int((<PeakBase>p_peak).mz * scaler)
            bin_b[index] += (<PeakBase>p_peak).intensity

        # Calculate dot product and normalizers in parallel
        n = k
        for i in prange(n, schedule='static', num_threads=4):
            a = bin_a[i]
            b = bin_b[i]
            n_a += a * a
            n_b += b * b
            z += a * b

        free(bin_a)
        free(bin_b)
        n_ab = sqrt(n_a) * sqrt(n_b)
        if n_ab == 0:
            return 0.0
        return z / n_ab


@cython.boundscheck(False)
cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):
    cdef:
        double hi, scaler
        PeakBase peak
        cnp.ndarray[double, ndim=1] bin_a, bin_b
        size_t i, n, index

    if peak_collection is PeakIndex:
        hi = 0
        peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is object:
        hi = 0
        for obj in peak_set_a:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
        for obj in peak_set_b:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
    elif peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
        peak = peak_set_a._mz_ordered[-1]
        hi = peak.mz
        peak = peak_set_b._mz_ordered[-1]
        hi = max(peak.mz, hi)
    else:
        hi = 0
        peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    scaler = 10 ** precision
    hi = (hi * scaler) + 1
    bin_a = zeros(<int>hi)
    bin_b = zeros(<int>hi)

    if peak_collection is object:
        n = len(peak_set_a)
    else:
        n = peak_set_a.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_a.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_a[i]
        else:
            peak = peak_set_a.getitem(i)
        index = int(peak.mz * scaler)
        bin_a[index] += peak.intensity

    if peak_collection is object:
        n = len(peak_set_b)
    else:
        n = peak_set_b.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_b.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_b[i]
        else:
            peak = peak_set_b.getitem(i)
        index = int(peak.mz * scaler)
        bin_b[index] += peak.intensity
    return (bin_a, bin_b)


@cython.boundscheck(False)
@cython.cdivision(True)
cdef double calculate_dot_product(cnp.ndarray[double, ndim=1] bin_a, cnp.ndarray[double, ndim=1] bin_b, bint normalize=True):
    cdef:
        double n_a, n_b, n_ab, z, a, b
        long i, n
    with nogil:
        n_a = 0
        n_b = 0
        z = 0
        n = bin_a.shape[0]
        for i in prange(n, schedule='static'):
            a = bin_a[i]
            b = bin_b[i]
            n_a += a * a
            n_b += b * b
            z += a * b
        if normalize:
            n_ab = sqrt(n_a) * sqrt(n_b)
            if n_ab == 0:
                return 0.0
            return z / n_ab
        return z



cdef class SpectrumAlignment(object):
    cdef:
        public list peak_pairs
        public double shift
        public double score
        public object peak_set_a
        public object peak_set_b
        public PeakType peak_type
        public double norm_aa_cosine
        public double norm_bb_cosine
        public bint normalize
        public bint sqrt_transform

    def __init__(self, peak_set_a, peak_set_b, double error_tolerance=2e-5, double shift=0.0,
                 bint normalize=True, bint sqrt_transform=False):
        self.peak_set_a = peak_set_a
        self.peak_set_b = peak_set_b
        self.peak_pairs = []
        self.shift = shift

        self.score = 0
        self.norm_aa_cosine = 1.0
        self.norm_bb_cosine = 1.0

        self.normalize = normalize
        self.sqrt_transform = sqrt_transform

        self._determine_peak_type()
        self.calculate_normalization()
        self.align(error_tolerance, shift)

    cpdef calculate_normalization(self):
        cdef:
            size_t i, n
            PeakSet peak_set
            DeconvolutedPeakSet dpeak_set
            double acc, intensity, acc_cosine
            bint sqrt_transform

        sqrt_transform = self.sqrt_transform
        if self.peak_type == PeakType.Fitted:
            peak_set = <PeakSet>self.peak_set_a
            n = peak_set.get_size()
            acc_cosine = 0.0
            for i in range(n):
                intensity = peak_set.getitem(i).intensity
                if sqrt_transform:
                    intensity = sqrt(intensity)
                acc_cosine += intensity ** 2
            self.norm_aa_cosine = sqrt(acc_cosine)
            peak_set = <PeakSet>self.peak_set_b
            n = peak_set.get_size()
            acc_cosine = 0.0
            for i in range(n):
                intensity = peak_set.getitem(i).intensity
                if sqrt_transform:
                    intensity = sqrt(intensity)
                acc_cosine += intensity ** 2
            self.norm_bb_cosine = sqrt(acc_cosine)
        elif self.peak_type == PeakType.Deconvoluted:
            dpeak_set = <DeconvolutedPeakSet>self.peak_set_a
            n = dpeak_set.get_size()
            acc_cosine = 0.0
            for i in range(n):
                intensity = dpeak_set.getitem(i).intensity
                if sqrt_transform:
                    intensity = sqrt(intensity)
                acc_cosine += intensity ** 2
            self.norm_aa_cosine = sqrt(acc_cosine)
            dpeak_set = <DeconvolutedPeakSet>self.peak_set_b
            n = dpeak_set.get_size()
            acc_cosine = 0.0
            for i in range(n):
                intensity = dpeak_set.getitem(i).intensity
                if sqrt_transform:
                    intensity = sqrt(intensity)
                acc_cosine += intensity ** 2
            self.norm_bb_cosine = sqrt(acc_cosine)

    cpdef _determine_peak_type(self):
        if isinstance(self.peak_set_a, DeconvolutedPeakSet):
            if isinstance(self.peak_set_b, DeconvolutedPeakSet):
                self.peak_type = PeakType.Deconvoluted
            else:
                raise TypeError(
                    "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
        elif isinstance(self.peak_set_a, PeakSet):
            if isinstance(self.peak_set_b, PeakSet):
                self.peak_type = PeakType.Fitted
            else:
                raise TypeError(
                    "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))
        else:
            raise TypeError(
                "Peak set types must match, got %s and %s" % (type(self.peak_set_a), type(self.peak_set_b)))

    cpdef align(self, double error_tolerance=2e-5, double shift=0.0):
        cdef:
            list pairs
        pairs = []
        if self.peak_type == PeakType.Fitted:
            pairs = convolve_peak_sets(<PeakSet>self.peak_set_a, <PeakSet>self.peak_set_b, error_tolerance)
            if shift != 0:
                pairs.extend(
                    convolve_peak_sets(
                        <PeakSet>self.peak_set_a,
                        <PeakSet>self.peak_set_b, error_tolerance, shift))
        elif self.peak_type == PeakType.Deconvoluted:
            pairs = convolve_deconvoluted_peak_sets(
                <DeconvolutedPeakSet>self.peak_set_a,
                <DeconvolutedPeakSet>self.peak_set_b, error_tolerance)
            if shift != 0:
                pairs.extend(
                    convolve_deconvoluted_peak_sets(
                        <DeconvolutedPeakSet>self.peak_set_a,
                        <DeconvolutedPeakSet>self.peak_set_b, error_tolerance, shift))
        self.peak_pairs = pairs
        self._calculate_score()

    cpdef _calculate_score(self):
        self.score = convolved_dot_product(
            self.peak_pairs, self.peak_type, sqrt_transform=self.sqrt_transform)
        if self.normalize:
            self.score /= (self.norm_aa_cosine * self.norm_bb_cosine)



cdef list convolve_peak_sets(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5, double shift=0.0):
    cdef:
        size_t i, n, j, m
        FittedPeak peak
        FittedPeak other
        tuple peaks_slice
        list peak_pairs
        list pairs_for_peak
        double pt

    peak_pairs = []
    n = peak_set_a.get_size()
    for i in range(n):
        pairs_for_peak = []
        peak = peak_set_a.getitem(i)
        pt = peak.mz + shift
        peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
        m = PyTuple_Size(peaks_slice)
        for j in range(m):
            other = <FittedPeak>PyTuple_GET_ITEM(peaks_slice, j)
            peak_pairs.append(PeakMatch._create(peak, other))
    return peak_pairs


cdef list convolve_deconvoluted_peak_sets(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,
                                          double error_tolerance=2e-5, double shift=0.0):
    cdef:
        size_t i, n, j, m
        DeconvolutedPeak peak
        DeconvolutedPeak other
        tuple peaks_slice
        list peak_pairs
        list pairs_for_peak
        double pt

    peak_pairs = []
    n = peak_set_a.get_size()
    for i in range(n):
        peak = peak_set_a.getitem(i)
        pt = peak.neutral_mass + shift
        peaks_slice = peak_set_b.all_peaks_for(pt, error_tolerance)
        m = PyTuple_Size(peaks_slice)
        for j in range(m):
            other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
            peak_pairs.append(PeakMatch._create(peak, other))
    return peak_pairs


cpdef double convolved_dot_product(list peak_pairs, PeakType peak_type=PeakType.Fitted, double norm_a=1.0,
                                   double norm_b=1.0, bint sqrt_transform=False):
    cdef:
        size_t i, n
        PeakBase peak, other
        list pairs
        PeakMatch pair
        double d
        set seen1, seen2
        object key1, key2
    seen1 = set()
    seen2 = set()
    d = 0.0
    n = PyList_Size(peak_pairs)
    for i in range(n):
        pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
        if sqrt_transform:
            pair.score = (sqrt(pair.peak_a.intensity) / norm_a) * (sqrt(pair.peak_b.intensity) / norm_b)
        else:
            pair.score = (pair.peak_a.intensity / norm_a) * (pair.peak_b.intensity / norm_b)
    PyList_Sort(peak_pairs)

    for i in range(n - 1, -1, -1):
        pair = <PeakMatch>PyList_GET_ITEM(peak_pairs, i)
        if peak_type == PeakType.Fitted:
            key1 = (<FittedPeak>pair.peak_a).peak_count
            if key1 in seen1:
                continue
            key2 = (<FittedPeak>pair.peak_b).peak_count
            if key2 in seen2:
                continue
        elif peak_type == PeakType.Deconvoluted:
            key1 = (<DeconvolutedPeak>pair.peak_a)._index.neutral_mass
            if key1 in seen1:
                continue
            key2 = (<DeconvolutedPeak>pair.peak_b)._index.neutral_mass
            if key2 in seen2:
                continue
        seen1.add(key1)
        seen2.add(key2)
        d += pair.score
    return d


cdef double convolve_peak_sets2_fitted(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m
        FittedPeak peak
        FittedPeak other
        PeakSet peaks_slice
        double acc

    acc = 0.0
    n = peak_set_a.get_size()
    for i in range(n):
        peak = peak_set_a.getitem(i)
        peaks_slice = peak_set_b._between(
            peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
        m = peaks_slice.get_size()
        for j in range(m):
            other = peaks_slice.getitem(j)
            acc += peak.intensity * other.intensity
    return acc


cdef double convolve_peak_sets2_deconvoluted(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b,
                                             double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m
        DeconvolutedPeak peak
        DeconvolutedPeak other
        tuple peaks_slice
        double acc

    acc = 0.0
    n = peak_set_a.get_size()
    for i in range(n):
        peak = peak_set_a.getitem(i)
        peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
        m = PyTuple_Size(peaks_slice)
        for j in range(m):
            other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
            acc += peak.intensity * other.intensity
    return acc


cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):
    cdef:
        double ab, aa, bb
    if peak_collection is PeakSet:
        ab = convolve_peak_sets2_fitted(peak_set_a, peak_set_b, error_tolerance)
        aa = convolve_peak_sets2_fitted(peak_set_a, peak_set_a, error_tolerance)
        bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)
    elif peak_collection is PeakIndex:
        ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)
        aa = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_a.peaks, error_tolerance)
        bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)
    elif peak_collection is DeconvolutedPeakSet:
        ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)
        aa = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_a, error_tolerance)
        bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)
    elif peak_collection is object:
        raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))

    return (ab) / (sqrt((aa)) * sqrt((bb)))


cpdef SpectrumAlignment align_peaks(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5,
                                    double shift=0.0, bint normalize=True):
    if peak_collection is PeakSet or peak_collection is DeconvolutedPeakSet or peak_collection is DeconvolutedPeakSetIndexed:
        return SpectrumAlignment(peak_set_a, peak_set_b, error_tolerance, shift, normalize, sqrt_transform=False)
    elif peak_collection is PeakIndex:
        return SpectrumAlignment(peak_set_a.peaks, peak_set_b.peaks, error_tolerance, shift, normalize, sqrt_transform=False)
    elif peak_collection is object:
        raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))



""" Waters
    MassLynx Python Chromatogram reader SDK
"""
import copy

import ctypes
from ctypes import*

from array import *

from .MassLynxRawReader import *
from .MassLynxRawDefs import MassLynxBaseType

class MassLynxRawChromatogramReader(MassLynxRawReader):
    """Read masslynx chromatogram data"""
    def __init__(self, source ):
        super(MassLynxRawChromatogramReader, self).__init__(source, MassLynxBaseType.CHROM)

    #@classmethod
    #def CreateFromPath( cls, path ):                      # alternative constructor - pass class to constructor
    #    return cls(MassLynxRawReader.fromPath( path, 3 ))                     # initalise with reader

    #@classmethod
    #def CreateFromReader( cls, sourceReader ):                  # alternative constructor - pass class to constructor
    #     return cls(MassLynxRawReader.fromReader( sourceReader, 3 ))                     # initalise with reader

    def ReadTIC( self, whichFunction ):
        times = []
        intensities = []

        # create the retrun values
        size = c_int(0)
        pTimes = c_void_p()
        pIntensities = c_void_p()

        # read tic
        readTIC = MassLynxRawReader.massLynxDll.readTICChromatogram
        readTIC.argtypes = [c_void_p, c_int, POINTER(c_void_p), POINTER(c_void_p), POINTER(c_int)]
        super(MassLynxRawChromatogramReader, self).CheckReturnCode( readTIC(self._getReader(),whichFunction, pTimes, pIntensities, size) )

        # fill the array
        pT = cast(pTimes,POINTER(c_float))
        pI = cast(pIntensities,POINTER(c_float))

        times = pT[0:size.value]
        intensities = pI[0:size.value]

        # dealocate memory
        MassLynxRawReader.ReleaseMemory( pTimes)
        MassLynxRawReader.ReleaseMemory( pIntensities)

        return times, intensities

    def ReadBPI( self, whichFunction ):
        #times = []
        #intensities = []
        #try:
        # create the retrun values
        size = c_int(0)
        pTimes = c_void_p()
        pIntensities = c_void_p()

        # read tic
        readBPI = MassLynxRawReader.massLynxDll.readTICChromatogram
        readBPI.argtypes = [c_void_p, c_int, POINTER(c_void_p), POINTER(c_void_p), POINTER(c_int)]
        super(MassLynxRawChromatogramReader, self).CheckReturnCode( readBPI(self._getReader(),whichFunction, pTimes, pIntensities, size) )

        # fill the array
        pT = cast(pTimes,POINTER(c_float))
        pI = cast(pIntensities,POINTER(c_float))

        times = pT[0:size.value]
        intensities = pI[0:size.value]

        # dealocate memory
        MassLynxRawReader.ReleaseMemory( pTimes)
        MassLynxRawReader.ReleaseMemory( pIntensities)

        #except MassLynxException as e:
        #    e.Handler()
        #    return [], []

        return times, intensities

    def ReadMassChromatogram( self, whichFunction, whichMass, massTollerance, daughters ):
        # just call multiple mass with list of 1
        whichMasses = [whichMass]
        times, intensities =  self.ReadMassChromatograms( whichFunction, whichMasses, massTollerance, daughters )

        return times, intensities[0]

    def ReadMassChromatograms( self, whichFunction, whichMasses, massTollerance, daughters ):
        times = []
        intensities = []

        # get the array of masses
        numMasses = len(whichMasses)
        masses = (c_float * numMasses)(*whichMasses)

        # create the retrun values
        size = c_int(0)
        pTimes = c_void_p()

        # create array of pointers to hold return intensities
        pIntensities = c_void_p()

        readMassChroms = MassLynxRawReader.massLynxDll.readMassChromatograms
        readMassChroms.argtypes = [c_void_p, c_int, POINTER(c_float), c_int, POINTER(c_void_p), POINTER(c_void_p), c_float, c_bool, POINTER(c_int)]
        super(MassLynxRawChromatogramReader, self).CheckReturnCode( readMassChroms( self._getReader(), whichFunction, masses, numMasses, pTimes, pIntensities, massTollerance, daughters, size))

        # fill the array and free memory
        pT = cast(pTimes,POINTER(c_float))
        times = pT[0:size.value]
        MassLynxRawReader.ReleaseMemory( pTimes)

        # fill in the mass chroms and free memory
        pI = cast(pIntensities ,POINTER(c_float))
        for index in range(0, numMasses ):
            intensities.append( pI[index * size.value :(index + 1)* size.value ])
#            MassLynxRawReader.ReleaseMemory( ppIntensities[ index] )
        MassLynxRawReader.ReleaseMemory( pIntensities )


        return times, intensities

   # def getMRMsinFunction( self, whichFunction ):
   #     try:
   #         # get the number of MRM transitions
   #         numberMRMs = 0
   #         getMRMsInFunction = RawReader.massLynxDll.getMRMsInFunction
   #         getMRMsInFunction.argtypes = [c_void_p, c_int, POINTER(c_int)]
   #         RawReader.CheckReturnCode(getMRMsInFunction(RawReader.getReader(self), whichFunction, numberMRMs))

   #     except MassLynxException as e:
   #         e.Handler()
   #         return 0

   #     return numberMRMs;

   # def readMRM( self, whichFunction, whichMRM ):
   #     try:
   #         # check we are requesting a valid MRM index - no...
   #         # client  code should not know about data
   ##         if ( self.getMRMsinFunction( whichFunction ) > whichMRM ):

   #         # create the float array
   #         scans = self.getScansInFunction(whichFunction)
   #         times = (c_float*scans)()
   #         intensities = (c_float*scans)()

   #         # check we have some scans ?

   #         readMRM = RawReader.massLynxDll.readMRMChromatogram
   #         readMRM.argtypes = [c_void_p, c_int, c_int, POINTER(c_float), POINTER(POINTER(c_float))]
   #         RawReader.CheckReturnCode( readMRM( RawReader.getReader(self), whichFunction, whichMRM, times, intensities)) # test with invalid MRM

   #     except MassLynxException as e:
   #         e.Handler()
   #         return [], []

   #     return list(times), list(intensities)


   # def readMRMs( self, whichFunction ):
   #     try:
   #         numberMRMs = self.getMRMsinFunction( whichFunction )

   #         # create the float arrays
   #         scans = self.getScansInFunction(whichFunction)
   #         times = (c_float*scans)()

   #         # create array of pointers to hold return intensities
   #         intensities = (POINTER(c_float) * numberMRMs)()
   #         for index in range(0, numberMRMs ):
   #             intensities[ index ] = (c_float * scans)()

   #         readMRMs = RawReader.massLynxDll.readMRMChromatograms
   #         readMRMs.argtypes = [c_void_p, c_int, c_int, POINTER(c_float), POINTER(POINTER(c_float))]
   #         RawReader.CheckReturnCode( readMRMs( RawReader.getReader(self), whichFunction, numberMRMs, times, intensities,))

   #     except MassLynxException as e:
   #         e.Handler()
   #         return [], []

   #     # create the return types
   #     intensities_lists = []
   #     for index in range(0, numberMRMs ):
   #         intensities_lists.append( intensities[index][0:scans] )

   #     return list(times), intensities_lists



cimport cython

from cpython.list cimport PyList_GetItem, PyList_Size

import numpy as np
cimport numpy as np

from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, SimpleInterval, IntervalTreeNode2D

from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature, LCMSFeatureTreeNode


cdef class FeatureGraphNode(SpanningMixin):

    def __init__(self, feature, index, edges=None):
        self.start = feature.start_time
        self.end = feature.end_time
        if edges is None:
            edges = set()
        self.feature = feature
        self.index = index
        self.edges = edges
        self.center = self._average_center()
        self.mz = feature.mz

    def as_arrays(self, dtype=float):
        return self.feature.as_arrays(dtype)

    cpdef double _average_center(self):
        cdef:
            double total, abundance, intensity
            LCMSFeatureTreeNode node
            size_t i, n

        total = 0
        abundance = 0
        n = self.feature.get_size()
        for i in range(n):
            node = self.feature.getitem(i)
            intensity = node.total_intensity()
            total += node.time * intensity
            abundance += intensity
        return total / abundance

    @property
    def start_time(self):
        return self.start

    @property
    def end_time(self):
        return self.end

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.feature,)

    def __index__(self):
        return self.index

    def __hash__(self):
        return hash(self.index)

    def __eq__(self, other):
        return self.feature == other.feature


cdef class DeconvolutedFeatureGraphNode(FeatureGraphNode):

    def __init__(self, feature, index, edges=None):
        super().__init__(feature, index, edges)
        self.neutral_mass = feature.neutral_mass
        self.charge = feature.charge


cdef class IonMobilityProfileFeatureGraphNode(DeconvolutedFeatureGraphNode):

    def __init__(self, feature, index, edges=None):
        super().__init__(feature, index, edges)
        self.ion_mobility_interval = feature.ion_mobility_interval


cdef class FeatureGraphEdge(object):
    def __init__(self, node_a, node_b, transition, weight=1, mass_error=0, rt_error=0):
        self.node_a = node_a
        self.node_b = node_b
        self.transition = transition
        self.weight = weight
        self.mass_error = mass_error
        self.rt_error = rt_error
        if self.node_a.index < self.node_b.index:
            self._hash = hash(((self.node_a.index, self.node_b.index)))
        else:
            self._hash = hash(((self.node_b.index, self.node_a.index)))
        self.node_a.edges.add(self)
        self.node_b.edges.add(self)

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__, self.node_a.feature, self.node_b.feature,
            self.transition)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self.node_a.index == other.node_a.index and self.node_b.index == other.node_b.index

    def __ne__(self, other):
        return not self == other

    cpdef remove(self):
        self.node_a.edges.remove(self)
        self.node_b.edges.remove(self)



cdef class TimeQuery(SpanningMixin):
    def __init__(self, feature: LCMSFeature, width: float = 0):
        self.start = feature.get_start_time() - width
        self.end = feature.get_end_time() + width

    def __repr__(self):
        return "TimeQuery(%f, %f)" % (self.start, self.end)


cdef class PPMQuery(SpanningMixin):
    def __init__(self, center, error_tolerance):
        self.start = center - center * error_tolerance
        self.end = center + center * error_tolerance

    def __repr__(self):
        return "PPMQuery(%f, %f)" % (self.start, self.end)


cdef (double, double) time_query(LCMSFeature x, double width):
    cdef:
        double out1, out2

    out1 = x.get_start_time() - width
    out2 = x.get_end_time() + width
    return out1, out2


cdef (double, double) ppm_query(double x, double ppm):
    cdef:
        double out1, out2

    out1 = x - x * ppm
    out2 = x + x * ppm
    return out1, out2


@cython.binding(True)
@cython.cdivision(True)
def find_edges(self, FeatureGraphNode node, double query_width=2., double error_tolerance=1.5e-5, **kwargs):
    cdef:
        list nodes
        set edges
        FeatureGraphNode match
        size_t i, n
        IntervalTreeNode2D rt_tree
        double[2] starts, ends
        (double, double) query, mass_query

    query = time_query(node.feature, query_width)
    mass_query = ppm_query(node.mz, error_tolerance)
    starts = [query[0], mass_query[0]]
    ends = [query[1], mass_query[1]]
    rt_tree = self.rt_tree
    nodes = rt_tree.overlaps_2d(starts, ends)

    edge_cls = self.edge_cls
    edges = <set>self.edges

    n = PyList_Size(nodes)
    for i in range(n):
        match = <FeatureGraphNode>PyList_GetItem(nodes, i)
        ppm_error = (node.mz - match.mz) / match.mz
        if abs(ppm_error) > error_tolerance:
            continue
        if match.index == node.index:
            continue
        rt_error = (node.center - match.center)

        edges.add(
            edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


@cython.binding(True)
@cython.cdivision(True)
def find_edges_neutral_mass(self, DeconvolutedFeatureGraphNode node, double query_width=2., double error_tolerance=1.5e-5, **kwargs):
    cdef:
        list nodes
        set edges
        IntervalTreeNode2D rt_tree
        DeconvolutedFeatureGraphNode match
        size_t i, n
        double[2] starts, ends
        (double, double) query, mass_query

    query = time_query(node.feature, query_width)
    mass_query = ppm_query(node.neutral_mass, error_tolerance)
    starts = [query[0], mass_query[0]]
    ends = [query[1], mass_query[1]]
    rt_tree = self.rt_tree
    nodes = rt_tree.overlaps_2d(starts, ends)
        # np.array([query.start, mass_query.start]), np.array([query.end, mass_query.end]))

    edge_cls = self.edge_cls
    edges = <set>self.edges

    n = PyList_Size(nodes)
    for i in range(n):
        match = <DeconvolutedFeatureGraphNode>PyList_GetItem(nodes, i)
        if match.charge != node.charge:
            continue
        ppm_error = (node.neutral_mass - match.neutral_mass) / match.neutral_mass
        if abs(ppm_error) > error_tolerance:
            continue
        if match.index == node.index:
            continue
        rt_error = (node.center - match.center)

        edges.add(
            edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


@cython.binding(True)
@cython.cdivision(True)
def find_edges_neutral_mass_ion_mobility(self, IonMobilityProfileFeatureGraphNode node, double query_width=2., double error_tolerance=1.5e-5, **kwargs):
    cdef:
        list nodes
        set edges
        IntervalTreeNode2D rt_tree
        IonMobilityProfileFeatureGraphNode match
        size_t i, n
        double[2] starts, ends
        (double, double) query, mass_query

    query = time_query(node.feature, query_width)
    mass_query = ppm_query(node.neutral_mass, error_tolerance)
    starts = [query[0], mass_query[0]]
    ends = [query[1], mass_query[1]]
    rt_tree = self.rt_tree
    nodes = rt_tree.overlaps_2d(starts, ends)

    edge_cls = self.edge_cls
    edges = <set>self.edges

    n = PyList_Size(nodes)
    for i in range(n):
        match = <IonMobilityProfileFeatureGraphNode>PyList_GetItem(nodes, i)
        if match.charge != node.charge:
            continue
        # If there isn't an overlap in the IM dimension, then they aren't the same molecule even
        # if charge, mass, and elution time do match
        if not match.ion_mobility_interval.overlaps(node.ion_mobility_interval):
            continue
        ppm_error = (node.neutral_mass - match.neutral_mass) / \
            match.neutral_mass
        if abs(ppm_error) > error_tolerance:
            continue
        if match.index == node.index:
            continue
        rt_error = (node.center - match.center)
        self.edges.add(
            edge_cls(
                node, match, None, mass_error=ppm_error, rt_error=rt_error))


@cython.binding(True)
cpdef list connected_components(self):
    cdef:
        set pool, current_component, visited
        list components
        FeatureGraphNode node
        FeatureGraphEdge edge
        int i, j

    pool = set(self.nodes)
    components = []

    i = 0
    while pool:
        i += 1
        current_component = set()
        visited = set()
        node = pool.pop()
        current_component.add(node)
        j = 0
        while current_component:
            j += 1
            node = current_component.pop()
            visited.add(node)
            for edge in node.edges:
                if edge.node_a not in visited and edge.node_a in pool:
                    current_component.add(edge.node_a)
                    pool.remove(edge.node_a)
                if edge.node_b not in visited and edge.node_b in pool:
                    current_component.add(edge.node_b)
                    pool.remove(edge.node_b)
        components.append(list(visited))
    return components


"""A Tk-based spectrum viewer that integrates :mod:`ms_deisotope`'s spectrum processing
features.
"""
import sys
import os

import array

import matplotlib as mpl

import click

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg,)
from matplotlib.widgets import SpanSelector
import numpy as np

import tkinter as tk
from tkinter import Tk
from tkinter import ttk

import tkinter.filedialog as tkfiledialog

import ms_deisotope
from ms_deisotope.data_source import ScanBunch, ChargeNotProvided
from ms_deisotope.peak_set import EnvelopePair
from ms_deisotope.plot import (draw_raw, draw_peaklist, annotate_isotopic_peaks)

averagine_label_map = {
    "peptide": ms_deisotope.peptide,
    "glycan": ms_deisotope.glycan,
    "glycopeptide": ms_deisotope.glycopeptide,
    "heparan sulfate": ms_deisotope.heparan_sulfate
}


def interactive_console(**kwargs):
    import code
    code.interact(local=kwargs)


class Cursor(object):
    def __init__(self, ax, binding):
        self.ax = ax
        self.binding = binding

    def mouse_move(self, event):
        if not event.inaxes:
            return
        x, y = event.xdata, event.ydata
        if y > 1e4:
            text = "m/z=%1.2f, inten=%1.2g" % (x, y)
        else:
            text = "m/z=%1.2f, inten=%1.2f" % (x, y)
        self.binding.set(text)


class SummaryChromatogramBuilder(object):
    def __init__(self):
        self.tic = array.array('d')
        self.bpc = array.array('d')
        self.time = array.array('d')

    def add(self, scan):
        self.time.append(scan.scan_time)
        self.tic.append(scan.arrays.intensity.sum())
        self.bpc.append(scan.arrays.intensity.max())

    def reset(self):
        self.tic = array.array('d')
        self.bpc = array.array('d')
        self.time = array.array('d')


if issubclass(ttk.Frame, object):
    _BaseFrame = ttk.Frame
else:
    class _BaseFrame(object, ttk.Frame):
        pass


class SpectrumViewer(_BaseFrame):
    def __init__(self, master, denoise=4, ms1_score_threshold=20.0):
        ttk.Frame.__init__(self, master)
        self.root = master
        self._ms_file_name = None
        self.reader = None
        self.scan = None
        self.denoise = denoise
        self.ms1_score_threshold = ms1_score_threshold
        self.draw_isotopic_patterns = True
        self.configure_toolbar()
        self.configure_canvas()
        self.configure_treeview()
        self.configure_display_row()
        self.populate()
        self.draw_plot()

    @property
    def ms_file_name(self):
        return self._ms_file_name

    @ms_file_name.setter
    def ms_file_name(self, value):
        if value not in (None, '') and os.path.exists(value):
            print("Loading %r" % value)
            self._ms_file_name = value
            self.reader = ms_deisotope.MSFileLoader(self.ms_file_name)
            self.populate()

    def set_ms_file(self, value, populate=True):
        self._ms_file_name = value
        self.reader = ms_deisotope.MSFileLoader(self.ms_file_name)
        if populate:
            self.populate()

    def select_ms_file(self):
        file_name = tkfiledialog.askopenfilename()
        if file_name is not None and file_name != '':
            self.ms_file_name = file_name

    def do_layout(self):
        self.grid(sticky=tk.N + tk.W + tk.E + tk.S)
        tk.Grid.rowconfigure(self, 0, weight=1)
        tk.Grid.columnconfigure(self, 0, weight=1)

    def configure_canvas(self):
        self.figure = Figure(dpi=100)
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.axis = self.figure.add_subplot(111)
        self.canvas.draw()
        canvas_widget = self.canvas.get_tk_widget()
        canvas_widget.grid(row=0, column=0, sticky=tk.N + tk.W + tk.E + tk.S)
        self.canvas_cursor = Cursor(self.axis, tk.StringVar(master=self.root))
        self.canvas.mpl_connect('motion_notify_event', self.canvas_cursor.mouse_move)
        self.span = SpanSelector(
            self.axis, self.zoom, 'horizontal', useblit=True,
            rectprops=dict(alpha=0.5, facecolor='red'))
        self.mz_span = None
        self.scan = None
        self.annotations = []
        self.canvas.draw()

    def configure_toolbar(self):
        self.toolbar = tk.Menu(self)
        self.toolbar.add_command(label='Open', command=self.select_ms_file)
        self.toolbar.add_command(label='Interact', command=self._interact)
        self.root.config(menu=self.toolbar)

    def _interact(self):
        local_vars = {
            "self": self,
            "scan": self.scan,
        }
        return interactive_console(**local_vars)

    def _zoom_in(self, min_mz, max_mz):
        arrays = self.scan.arrays
        subset = arrays.between_mz(min_mz, max_mz).intensity
        if len(subset) > 0 and subset.max() > 0 and subset.min() < subset.max():
            most_intense = np.max(subset)
            y_cap = most_intense * 1.2
            delta = max_mz - min_mz
            if delta > 20:
                max_mz += min(delta * 0.1, 100)
                min_mz -= min(delta * 0.1, 100)
            self.mz_span = (min_mz, max_mz)
            self.axis.set_xlim(min_mz, max_mz)
            self.axis.set_ylim(0, y_cap)
            self.annotate_plot(min_mz, max_mz)
            self.figure.canvas.draw()

    def zoom(self, xmin, xmax):
        # If the bounds are not close, we're zooming in
        if not self.scan:
            return
        arrays = self.scan.arrays
        # Default zoom
        if (xmax - xmin) <= .1:
            min_peak = 0
            max_peak = len(arrays[0]) - 1
            self.mz_span = None
            xmin, xmax = arrays.mz[min_peak], arrays.mz[max_peak]
            xmin = max(xmin - 100, 0)
        if (xmin - 50) < 0:
            if xmin > -20:
                xmin = -20
        self._zoom_in(xmin, xmax)

    def clear_annotations(self):
        for anno in self.annotations:
            try:
                anno.remove()
            except ValueError:
                break
        self.annotations = []

    def annotate_plot(self, min_mz=None, max_mz=None):
        self.clear_annotations()
        if min_mz is None:
            min_mz = 0
        if max_mz is None:
            max_mz = self.scan.arrays.mz[-1]
        subset = self.scan.deconvoluted_peak_set.between(min_mz, max_mz, use_mz=True)
        if not subset:
            return
        threshold = 0.0
        threshold_list = ([max(i.intensity for i in p.envelope) for p in subset])
        if threshold_list:
            threshold = np.mean(threshold_list)
        threshold_list = ([max(i.intensity for i in p.envelope) for p in subset
                           if max(i.intensity for i in p.envelope) > threshold])
        if threshold_list:
            threshold = np.mean(threshold_list)
        for peak in subset:
            if peak.intensity > threshold:
                label = '%0.2f (%d)' % (peak.neutral_mass, peak.charge)
                pt = max(peak.envelope, key=lambda x: x.intensity)
                y = pt.intensity * 1.05
                x = np.average(
                    [p.mz for p in peak.envelope],
                    weights=[p.intensity for p in peak.envelope])
                self.annotations.append(
                    self.axis.text(x, y, label, ha='center', clip_on=True, fontsize=10))

    def _process_scan(self):
        self.scan.pick_peaks(signal_to_noise_threshold=1.5)
        if self.scan.ms_level == 1:
            averagine_value = self.ms1_averagine_combobox.get()
            averagine_value = averagine_label_map[averagine_value]
            truncate_to = 0.95
            scorer = ms_deisotope.PenalizedMSDeconVFitter(self.ms1_score_threshold, 2.)
        else:
            averagine_value = self.msn_averagine_combobox.get()
            averagine_value = averagine_label_map[averagine_value]
            truncate_to = 0.8
            scorer = ms_deisotope.MSDeconVFitter(10.)
        max_charge_state = self.max_charge_state_var.get()
        min_charge_state = self.min_charge_state_var.get()
        if self.scan.ms_level > 1:
            prec_charge = self.scan.precursor_information.charge
            if prec_charge and prec_charge != ChargeNotProvided:
                max_charge_state = prec_charge
                if prec_charge < 0:
                    min_charge_state = -abs(min_charge_state)
        self.scan.deconvolute(
            averagine=averagine_value,
            scorer=scorer,
            truncate_after=1 - 1e-4,
            incremental_truncation=truncate_to,
            charge_range=(min_charge_state, max_charge_state))

    def draw_plot(self, scan=None, children=None):
        if children is None:
            children = []
        if scan is None or scan.arrays.mz.shape[0] == 0:
            return
        self.axis.clear()
        self.scan = scan
        self.children_scans = children
        if scan.ms_level == 1:
            averaging_level = self.ms1_scan_averaging_var.get()
            if averaging_level > 0:
                print("Averaging")
                if scan.arrays.mz.shape[0] > 1 :
                    self.scan = scan = scan.average(averaging_level)
            print("Denoising")
            self.scan = scan.denoise(self.denoise) if self.denoise else scan
        print("Processing Peaks")
        self._process_scan()
        print("Begin Drawing")
        scan = self.scan
        if scan.is_profile:
            print("Drawing Profile")
            draw_raw(*scan.arrays, ax=self.axis, color='black', lw=0.75)
            # lock in axis dimensions for profile spectra according to the raw profile
            self.axis.set_xlim(0, max(self.axis.get_xlim()))
        if scan.peak_set:
            print("Drawing Centroids")
            draw_peaklist(scan.peak_set, ax=self.axis, alpha=0.5, lw=1, color='grey', linestyle='--')
            if not scan.is_profile:
                # lock in axis dimensions for centroid spectra according to the picked peaks
                self.axis.set_xlim(0, max(self.axis.get_xlim()) + 2)
        if scan.deconvoluted_peak_set:
            print("Drawing Isotopic Envelope")
            draw_peaklist(
                [i for p in scan.deconvoluted_peak_set for i in p.envelope],
                ax=self.axis, alpha=0.6, lw=0.5, color='orange')
            draw_peaklist(
                [p.envelope[0] for p in scan.deconvoluted_peak_set if p.envelope[0].intensity > 0],
                ax=self.axis, alpha=0.6, lw=1,
                color='red')

        # draw isolation window and instrument reported precursor
        if children:
            ylim = scan.arrays.intensity.max()
            for child in children:
                if child.isolation_window and not child.isolation_window.is_empty():
                    self.axis.vlines(
                        [child.isolation_window.lower_bound, child.isolation_window.upper_bound],
                        0, ylim, linestyle='--', color='skyblue', lw=1)
                self.axis.vlines(child.precursor_information.mz, 0, ylim,
                                 linestyle='--', color='black', lw=1)
        if self.draw_isotopic_patterns:
            annotate_isotopic_peaks(self.scan, self.axis)
        if self.scan.precursor_information:
            ylim = scan.arrays.intensity.max()
            self.axis.vlines(self.scan.precursor_information.mz, 0, ylim, linestyle='-.', color='orange')
        if self.mz_span is not None:
            self._zoom_in(*self.mz_span)
        else:
            self.annotate_plot(None, None)
        self.canvas.draw()

    def on_row_click(self, event):
        selection = self.treeview.focus()
        item = self.treeview.item(selection)
        index = item['text']

        if self.reader is not None:
            scan = self.reader.get_scan_by_index(index)
            if scan.ms_level == 1:
                bunch = next(self.reader.start_from_scan(scan.id))
                if not isinstance(bunch, ScanBunch):
                    children = []
                else:
                    children = bunch.products
            else:
                children = []
            self.draw_plot(scan, children)
        else:
            self.draw_plot(None)

    def configure_display_row(self):
        self.display_row = ttk.Frame(self)
        self.display_row.grid(row=1, column=0, sticky=tk.W + tk.S + tk.E)
        self.cursor_label = ttk.Label(self.display_row, text=" " * 20)
        self.cursor_label.grid(row=0, padx=(10, 10))

        def update_label(*args, **kwargs):
            self.cursor_label['text'] = self.canvas_cursor.binding.get()

        self.canvas_cursor.binding.trace('w', update_label)

        self.ms1_averagine_combobox = ttk.Combobox(self.display_row, values=[
            "peptide",
            "glycan",
            "glycopeptide",
            "heparan sulfate",
        ], width=15)
        self.ms1_averagine_combobox.set("glycopeptide")
        self.ms1_averagine_combobox_label = ttk.Label(self.display_row, text="MS1 Averagine:")
        self.ms1_averagine_combobox_label.grid(row=0, column=1, padx=(10, 0))
        self.ms1_averagine_combobox.grid(row=0, column=2, padx=(1, 10))

        self.msn_averagine_combobox = ttk.Combobox(self.display_row, values=[
            "peptide",
            "glycan",
            "glycopeptide",
            "heparan sulfate",
        ], width=15)
        self.msn_averagine_combobox.set("peptide")
        self.msn_averagine_combobox_label = ttk.Label(self.display_row, text="MSn Averagine:")
        self.msn_averagine_combobox_label.grid(row=0, column=3, padx=(10, 0))
        self.msn_averagine_combobox.grid(row=0, column=4, padx=(1, 10))

        self.ms1_scan_averaging_label = ttk.Label(
            self.display_row, text="MS1 Signal Averaging:")
        self.ms1_scan_averaging_label.grid(row=0, column=5, padx=(10, 0))
        self.ms1_scan_averaging_var = tk.IntVar(self, value=2)
        self.ms1_scan_averaging = ttk.Entry(self.display_row, width=3)
        self.ms1_scan_averaging['textvariable'] = self.ms1_scan_averaging_var
        self.ms1_scan_averaging.grid(row=0, column=6, padx=(1, 3))

        self.min_charge_state_var = tk.IntVar(self, value=1)
        self.max_charge_state_var = tk.IntVar(self, value=12)
        self.min_charge_state_label = ttk.Label(
            self.display_row, text="Min Charge:")
        self.min_charge_state_label.grid(row=0, column=7, padx=(5, 0))
        self.min_charge_state = ttk.Entry(self.display_row, width=3)
        self.min_charge_state['textvariable'] = self.min_charge_state_var
        self.min_charge_state.grid(row=0, column=8, padx=(1, 3))

        self.max_charge_state_label = ttk.Label(
            self.display_row, text="Max Charge:")
        self.max_charge_state_label.grid(row=0, column=9, padx=(5, 0))
        self.max_charge_state = ttk.Entry(self.display_row, width=3)
        self.max_charge_state['textvariable'] = self.max_charge_state_var
        self.max_charge_state.grid(row=0, column=10, padx=(1, 3))

    def configure_treeview(self):
        self.treeview = ttk.Treeview(self)
        self.treeview['columns'] = ["id", "time", 'ms_level', 'precursor_mz', 'precursor_charge', 'activation']
        self.treeview.grid(row=2, column=0, sticky=tk.S + tk.W + tk.E + tk.N)

        self.treeview_scrollbar = ttk.Scrollbar(self, orient="vertical", command=self.treeview.yview)
        self.treeview_scrollbar.grid(row=2, column=0, sticky=tk.S + tk.E + tk.N)
        self.treeview.configure(yscrollcommand=self.treeview_scrollbar.set)

        self.treeview.heading('id', text="Scan ID")
        self.treeview.heading('#0', text='Index')
        self.treeview.heading("time", text='Time (min)')
        self.treeview.heading("ms_level", text='MS Level')
        self.treeview.heading("precursor_mz", text='Precursor M/Z')
        self.treeview.heading("precursor_charge", text='Precursor Z')
        self.treeview.heading("activation", text='Activation')
        self.treeview.column("#0", width=75)
        self.treeview.column("ms_level", width=75)
        self.treeview.column("time", width=75)
        self.treeview.column("precursor_mz", width=100)
        self.treeview.column("precursor_charge", width=100)
        self.treeview.bind("<<TreeviewSelect>>", self.on_row_click)

    def clear_treeview(self):
        children = self.treeview.get_children()
        if children:
            self.treeview.delete(*children)

    def _populate_range(self, start, stop):
        scans = []
        ended = False
        for i in range(start, stop):
            try:
                scan = self.reader[i]
            except Exception:
                ended = True
                break
            if scan.index % 5000 == 0:
                print(scan)
            i = scan.index
            values = [scan.id, "%0.4f" % scan.scan_time, scan.ms_level]
            if scan.ms_level > 1:
                values.extend([scan.precursor_information.mz, scan.precursor_information.charge,
                               str(scan.activation)])
            else:
                values.extend(['-', '-', '-'])
            scans.append(values)
        i = start
        for values in scans:
            self.treeview.insert('', 'end', values=values, text=i)
            i += 1
        if not ended:
            self.after(100, self._populate_range, stop, stop + 500)

    def populate(self, clear=True):
        if clear:
            self.clear_treeview()
        if self.reader is not None:
            self.reader.make_iterator(grouped=False)
            self.after(10, self._populate_range, 0, 500)


@click.command("ms-view")
@click.argument("filename", type=click.Path(exists=True, readable=True), required=False)
@click.option("-d", "--denoise", type=float, default=5.0)
def main(filename=None, denoise=5, ms1_score=20.0, msn_score=10.0, ms1_truncation=0.95, msn_truncation=0.8):
    base = Tk()
    base.title("ms_deisotope Spectrum Viewer")
    tk.Grid.rowconfigure(base, 0, weight=1)
    tk.Grid.columnconfigure(base, 0, weight=1)
    app = SpectrumViewer(base, denoise=denoise)
    app.do_layout()
    app.ms_file_name = filename

    app.mainloop()


if __name__ == '__main__':
    main.main()



import logging

from collections import defaultdict
from typing import DefaultDict, Dict, Iterator, List, Sized, Union
from ms_deisotope.data_source.metadata.activation import ActivationInformation
from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame, IonMobilitySourceRandomAccessFrameSource
from ms_deisotope.data_source.scan.scan import ProcessedScan, WrappedScan
from ms_deisotope.data_source.metadata.scan_traits import ion_mobility_drift_time

import numpy as np

from ms_deisotope.task import LogUtilsMixin
from ms_deisotope.data_source.scan.base import ScanBase, ScanBunch

from ms_deisotope.peak_set import IonMobilityProfileDeconvolutedPeakSolution, DeconvolutedPeakSet, IonMobilityDeconvolutedPeak
from ms_deisotope.peak_dependency_network.intervals import SpanningMixin

from ms_deisotope.data_source import PrecursorInformation, make_scan

from ms_deisotope.feature_map.profile_transform import sliding_mean

from ms_deisotope.feature_map.feature_fit import (
    IonMobilityProfileDeconvolutedLCMSFeature,
    DeconvolutedLCMSFeature)

from ms_deisotope.feature_map.feature_graph import (
    FeatureGraphEdge,
    IonMobilityProfileDeconvolutedFeatureGraph,
    IonMobilityProfileFeatureGraphNode)


from ms_deisotope._c.utils import correlation
from ms_deisotope._c.feature_map.profile_transform import interpolate_point

logger = logging.getLogger(__name__)


class PseudoXIC(object):
    __slots__ = ("x", "y", "source", )

    x: np.ndarray
    y: np.ndarray
    source: IonMobilityProfileDeconvolutedLCMSFeature

    def __init__(self, x: np.ndarray, y: np.ndarray, source: IonMobilityProfileDeconvolutedLCMSFeature):
        self.x = x
        self.y = y
        self.source = source

    def smooth(self) -> "PseudoXIC":
        return self.__class__(self.x, sliding_mean(self.y), self.source)

    def interpolate(self, t: float) -> float:
        return interpolate_point(self.x, self.y, t)

    @classmethod
    def from_feature(cls, source: IonMobilityProfileDeconvolutedLCMSFeature, smooth: bool = True) -> 'PseudoXIC':
        x, y = source.as_arrays(np.float32)
        if smooth:
            y = sliding_mean(y)
        return cls(x, y, source)

    def as_arrays(self):
        return self.x, self.y

    @property
    def neutral_mass(self) -> float:
        return self.source.neutral_mass

    @property
    def charge(self) -> int:
        return self.source.charge

    @property
    def mz(self) -> float:
        return self.source.mz

    @property
    def start_time(self) -> float:
        return self.x[0]

    @property
    def end_time(self) -> float:
        return self.x[-1]


def interpolate_feature(f: DeconvolutedLCMSFeature, Xp: np.ndarray) -> np.ndarray:
    x, y = f.as_arrays(np.float32)
    return np.interp(Xp, x, y, 0, 0)


def features_to_aligned_arrays(f1: DeconvolutedLCMSFeature, f2: DeconvolutedLCMSFeature, points_per_minute: int=150):
    start = f1.start_time
    end = f1.end_time
    n_points = int((end - start) * points_per_minute)
    X = np.linspace(start, end, n_points)
    Y1 = interpolate_feature(f1, X)
    Y2 = interpolate_feature(f2, X)
    return X, Y1, Y2


def reduce_feature(feature: IonMobilityProfileDeconvolutedLCMSFeature,
                   match: IonMobilityProfileDeconvolutedLCMSFeature,
                   include_mz: bool = True, include_envelope: bool = True) -> IonMobilityProfileDeconvolutedLCMSFeature:
    submatch = match.between_ion_mobilities(
        feature.ion_mobility_interval.start,
        feature.ion_mobility_interval.end,
        feature.start_time,
        feature.end_time,
        include_mz=include_mz,
        include_envelope=include_envelope)
    return submatch


class PrecursorProductCorrelationGraph(LogUtilsMixin):
    edge_cls = FeatureGraphEdge

    precursor_graph: IonMobilityProfileDeconvolutedFeatureGraph
    product_graph: IonMobilityProfileDeconvolutedFeatureGraph
    max_edge_count_per_node: int
    edges: set
    _is_built: bool

    def __init__(self, precursor_graph, product_graph, max_edge_count_per_node=1500):
        self.precursor_graph = precursor_graph
        self.product_graph = product_graph
        self.max_edge_count_per_node = max_edge_count_per_node
        self.edges = set()
        self._is_built = False

    def score_edge(self, feature: IonMobilityProfileDeconvolutedLCMSFeature,
                   match: IonMobilityProfileDeconvolutedLCMSFeature,
                   minimum_overlap_size: int=2) -> float:
        submatch: IonMobilityProfileDeconvolutedLCMSFeature = match.between_ion_mobilities(
            feature.ion_mobility_interval.start,
            feature.ion_mobility_interval.end,
            feature.start_time, feature.end_time,
            include_envelope=False,
            include_mz=False)
        if len(submatch) < minimum_overlap_size:
            return -1.0
        _x, y1, y2 = features_to_aligned_arrays(feature, submatch)
        corr = correlation(
            sliding_mean(y1),
            sliding_mean(y2),
        )
        if np.isnan(corr):
            corr = -1.0
        return corr

    def find_edges(self, precursor_node: IonMobilityProfileDeconvolutedFeatureGraph.node_cls, minimum_score=0.1, **kwargs):
        nodes = self.product_graph.rt_tree.overlaps_2d(
            np.array([precursor_node.start_time, 0.0]),
            np.array([precursor_node.end_time, precursor_node.neutral_mass])
        )
        charge = precursor_node.charge
        edges = set()
        for match in nodes:
            if match.charge > charge:
                continue
            if not match.ion_mobility_interval.overlaps(precursor_node.ion_mobility_interval):
                continue
            score = self.score_edge(precursor_node.feature, match.feature)
            if score < minimum_score:
                continue
            edge = self.edge_cls(
                precursor_node, match, score)
            edges.add(edge)
        edges = sorted(edges, key=lambda x: x.transition, reverse=True)
        keepers = edges[:self.max_edge_count_per_node]
        rest = edges[self.max_edge_count_per_node:]
        for edge in rest:
            edge.node_a.edges.remove(edge)
            edge.node_b.edges.remove(edge)
        self.edges.update(keepers)


    def build(self, min_charge=2, min_size=3):
        n = len(self.precursor_graph)
        last_logged = 0
        batch_size = 500
        for i, precursor in enumerate(self.precursor_graph):
            if precursor.charge < min_charge:
                continue
            if len(precursor.feature) < min_size:
                continue
            if i % 500 == 0 and i or i - last_logged >= batch_size:
                last_logged = i
                self.log(f"... Processed {i}/{n} precursors ({i / n * 100.0:0.2f}%)")
            self.find_edges(precursor)
        self._is_built = True

    def iterspectra(self, weight_scale_factor: float=1.0, **kwargs):
        if not self._is_built:
            self.build(**kwargs)
        iterator = PrecursorProductCorrelatingIterator(self.precursor_graph, self.product_graph, weight_scale_factor=weight_scale_factor)
        for batch in iterator:
            yield batch


PrecursorProductCorrelationGraph.log_with_logger(logger)


def edge_to_pseudopeak(edge: FeatureGraphEdge) -> IonMobilityProfileDeconvolutedPeakSolution:
    precursor = edge.node_a.feature
    product = edge.node_b.feature
    sub_product = reduce_feature(precursor, product)
    peak = IonMobilityProfileDeconvolutedPeakSolution.from_feature(sub_product)
    peak.intensity *= edge.transition ** 2
    return peak


def get_weighted_output_xic(edge: FeatureGraphEdge, rt: float, return_weight=False, weight_scale_factor: float=1.0):
    weights = []
    index = None
    o: FeatureGraphEdge
    for i, o in enumerate(edge.node_b.edges):
        if o.node_a.index == edge.node_a.index:
            index = i
        f = reduce_feature(edge.node_a.feature, o.node_a.feature)
        if not f:
            weights.append(0)
            continue
        f = PseudoXIC.from_feature(f)
        v = f.interpolate(rt)
        weights.append(v * o.transition)
    weights = np.array(weights)
    weights /= weights.max()
    if weight_scale_factor != 1.0:
        weights *= weight_scale_factor
    product = reduce_feature(edge.node_a.feature, edge.node_b.feature)
    f = PseudoXIC.from_feature(product)
    if return_weight:
        return (weights[index], edge.transition ** 2, weights[index] * f.interpolate(rt) * edge.transition ** 2, )
    return (weights[index] * f.interpolate(rt) * edge.transition ** 2)


def edge_at_time_to_pseudopeak(edge: FeatureGraphEdge, rt: float):
    precursor = edge.node_a.feature
    product = edge.node_b.feature
    sub_product = reduce_feature(precursor, product)
    peak = IonMobilityProfileDeconvolutedPeakSolution.from_feature(sub_product)
    peak.intensity = get_weighted_output_xic(edge, rt)
    return peak


class EdgeToXICCache(object):
    store: Dict[FeatureGraphEdge, PseudoXIC]
    ref_count: DefaultDict[FeatureGraphEdge, int]

    def __init__(self):
        self.store = dict()
        self.ref_count = defaultdict(int)

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value

    def __contains__(self, key):
        return key in self.store

    def __len__(self):
        return len(self.store)

    def checkout(self, key: FeatureGraphEdge):
        value = self.store[key]
        self.ref_count[key] += 1
        return value

    def checkin(self, key: FeatureGraphEdge):
        # Ref-count may be negative if someone
        # belatedly checks in an edge that has
        # already been evicted directly.
        count = self.ref_count[key]
        count -= 1
        if count <= 0:
            self.store.pop(key, None)
            self.ref_count.pop(key)
            return True
        else:
            self.ref_count[key] = count
        return False

    def __delitem__(self, key):
        self.store.pop(key, None)
        self.ref_count.pop(key, None)

    def clear(self):
        self.store.clear()
        self.ref_count.clear()

    def session(self) -> 'EdgeToXICCacheSession':
        return EdgeToXICCacheSession(self)


class EdgeToXICCacheSession(object):
    cache: EdgeToXICCache
    store: Dict[FeatureGraphEdge, PseudoXIC]

    def __init__(self, cache: EdgeToXICCache):
        self.cache = cache
        self.store = {}

    def __contains__(self, key):
        return key in self.cache

    def __getitem__(self, key: FeatureGraphEdge) -> PseudoXIC:
        try:
            return self.store[key]
        except KeyError:
            value = self.cache.checkout(key)
            self.store[key] = value
            return value

    def __setitem__(self, key, value):
        self.cache[key] = value
        self.store[key] = self.cache.checkout(key)

    def __delitem__(self, key):
        del self.store[key]
        self.cache.checkin(key)

    def remove(self, key: FeatureGraphEdge):
        if key in self.store:
            del self[key]

    def clear(self):
        for key in self.store:
            self.cache.checkin(key)
        self.store.clear()

    def __del__(self):
        self.clear()

    def __len__(self):
        return len(self.store)


class XICPseudoSpectrumGenerator(SpanningMixin):
    __slots__ = ('node', 'edge_to_pseudo_xic', 'edge_to_pseudo_xic_children', 'minimum_intensity', 'precursor_information')

    node: IonMobilityProfileFeatureGraphNode
    minimum_intensity: float
    precursor_information: PrecursorInformation
    edge_to_pseudo_xic: Union[EdgeToXICCacheSession, Dict[FeatureGraphEdge, PseudoXIC]]
    edge_to_pseudo_xic_children: Union[EdgeToXICCacheSession, Dict[FeatureGraphEdge, PseudoXIC]]

    def __init__(self, node: IonMobilityProfileFeatureGraphNode, minimum_intensity: float = 1, edge_to_pseudo_xic=None,
                 edge_to_pseudo_xic_children=None):
        if edge_to_pseudo_xic is None:
            edge_to_pseudo_xic = {}
        if edge_to_pseudo_xic_children is None:
            edge_to_pseudo_xic_children = {}

        self.node = node
        self.edge_to_pseudo_xic = edge_to_pseudo_xic
        self.edge_to_pseudo_xic_children = edge_to_pseudo_xic_children
        self.minimum_intensity = minimum_intensity
        self.start = self.node.start
        self.end = self.node.end
        self.precursor_information = None
        self._make_precursor_information()

    def _make_precursor_information(self):
        self.precursor_information = PrecursorInformation(
            self.node.feature.mz,
            self.node.feature.intensity,
            self.node.charge)

        dt = (self.node.ion_mobility_interval.end +
              self.node.ion_mobility_interval.start) / 2
        self.precursor_information._ion_mobility.add_ion_mobility(
            'ion mobility drift time', dt)

    def get_xic_for_edge(self, edge: FeatureGraphEdge) -> PseudoXIC:
        try:
            return self.edge_to_pseudo_xic[edge]
        except KeyError:
            f = reduce_feature(self.node.feature, edge.node_a.feature, False, False)
            if f:
                xic = PseudoXIC.from_feature(f)
            else:
                xic = None
            self.edge_to_pseudo_xic[edge] = xic
            return xic

    def get_xic_for_edge_child(self, edge: FeatureGraphEdge) -> PseudoXIC:
        try:
            return self.edge_to_pseudo_xic_children[edge]
        except KeyError:
            # Need to keep envelope and m/z as the product XIC's source will be used
            # to generate pseudo-peaks
            f = reduce_feature(self.node.feature, edge.node_b.feature, True, True)
            if f:
                xic = PseudoXIC.from_feature(f)
            else:
                xic = None
            self.edge_to_pseudo_xic_children[edge] = xic
            return xic

    def clear(self):
        self.edge_to_pseudo_xic.clear()
        self.edge_to_pseudo_xic_children.clear()

    def get_weighted_output(self, edge: FeatureGraphEdge, time: float, weight_scale_factor: float=1.0) -> float:
        weights = []
        index = None
        o: FeatureGraphEdge
        step_back = 0
        for i, o in enumerate(edge.node_b.edges):
            if o.node_a.index == self.node.index:
                index = i
            # If a parent node of a sharer of `edge` here doesn't span `time`,
            # evict it from the session, and if the index of `self.node` hasn't
            # been found yet, increment the backtrack counter by one before
            # skipping the interpolation.
            if not o.node_a.contains(time):
                self.edge_to_pseudo_xic.remove(o)
                if index is None:
                    step_back += 1
                continue
            f = self.get_xic_for_edge(o)
            if f is None:
                weights.append(0)
            else:
                v = f.interpolate(time)
                weights.append(v * o.transition ** 2)
        index -= step_back
        weights = np.array(weights)
        weights /= weights.max()
        if weight_scale_factor != 1.0:
            weights *= weight_scale_factor
        product = self.get_xic_for_edge_child(edge)
        return (weights[index] * product.interpolate(time) * edge.transition ** 2)

    def pseudopeak_for(self, edge: FeatureGraphEdge, time: float, weight_scale_factor: float=1.0) -> IonMobilityProfileDeconvolutedPeakSolution:
        weight = self.get_weighted_output(
            edge, time, weight_scale_factor=weight_scale_factor)
        xic = self.get_xic_for_edge_child(edge)
        f = xic.source
        peak = IonMobilityProfileDeconvolutedPeakSolution.from_feature(f)
        peak.intensity = weight
        return peak

    def pseudospectrum_for(self, time: float, weight_scale_factor: float = 1.0) -> DeconvolutedPeakSet:
        peaks = []
        for edge in self.node.edges:
            p = self.pseudopeak_for(
                edge, time, weight_scale_factor=weight_scale_factor)
            if p.intensity > self.minimum_intensity:
                peaks.append(p)
        peaks = DeconvolutedPeakSet(peaks)
        peaks.reindex()
        return peaks


class PrecursorProductCorrelatingIterator(LogUtilsMixin):
    precursor_graph: IonMobilityProfileDeconvolutedFeatureGraph
    product_graph: IonMobilityProfileDeconvolutedFeatureGraph
    weight_scale_factor: float

    edge_to_pseudo_xic: EdgeToXICCache
    edge_to_pseudo_xic_children: EdgeToXICCache

    def __init__(self, precursor_graph, product_graph, weight_scale_factor: float=1.0):
        self.precursor_graph = precursor_graph
        self.product_graph = product_graph
        self.scan_counter = 0
        self.ms1_times = None
        self.delta_ms1_time = None
        self.edge_to_pseudo_xic = EdgeToXICCache()
        self.edge_to_pseudo_xic_children = EdgeToXICCache()
        self._prepass_ms1()
        self.weight_scale_factor = weight_scale_factor

    def _prepass_ms1(self):
        times = set()
        for node in self.precursor_graph:
            node.generator = XICPseudoSpectrumGenerator(
                node, 1, self.edge_to_pseudo_xic.session(),
                self.edge_to_pseudo_xic_children.session())
            times.update(node.feature.times)
        self.ms1_times = np.array(sorted(times))
        self.delta_ms1_time = np.mean(np.diff(self.ms1_times))

    def __iter__(self):
        i = 0

        nodes_to_clear = []
        for i in range(len(self.ms1_times)):
            time = self.ms1_times[i]
            try:
                next_time = self.ms1_times[i + 1]
            except IndexError:
                next_time = time + self.delta_ms1_time

            peaks_for_scan = []
            msn_spectra = []

            for node_k in nodes_to_clear:
                node_k.generator.clear()
                node_k.generator = None
                for edge in list(node_k.edges):
                    del self.edge_to_pseudo_xic[edge]
                    edge.remove()

            nodes_to_clear = []
            self.log(
                f"... Processing time point {time:0.3f} with {self.scan_counter} scans produced, set size: \
{len(self.edge_to_pseudo_xic) + len(self.edge_to_pseudo_xic_children)} (\
{i / len(self.ms1_times) * 100.0:0.2f}%)")

            for node_k in self.precursor_graph.rt_tree.contains_point(time):
                cn, _ = node_k.feature.find_time(time)
                if not cn:
                    continue
                peaks_for_scan.extend(cn.members)
                generator: XICPseudoSpectrumGenerator = node_k.generator
                msn_for = generator.pseudospectrum_for(
                    time, weight_scale_factor=self.weight_scale_factor)
                if msn_for:
                    pinfo = generator.precursor_information.copy()
                    pinfo.intensity = cn.total_intensity()
                    msn_spectra.append(
                        (msn_for, pinfo)
                    )
                # Need to check if this is the last time point for this precursor node
                # so it can have its generator destroyed.
                if abs(node_k.end - time) < 1e-3:
                    nodes_to_clear.append(node_k)

            msn_spectra = [(t, ) + p for t, p in zip(np.linspace(time,
                                                                next_time, len(msn_spectra) + 2)[1:-1], msn_spectra)]
            peaks_for_scan = DeconvolutedPeakSet(peaks_for_scan)
            peaks_for_scan.reindex()

            ms1_scan = make_scan(
                id=f"merged={self.scan_counter}",
                ms_level=1, scan_time=time,
                index=self.scan_counter,
                deconvoluted_peak_set=peaks_for_scan)

            self.scan_counter += 1
            msn_scans = []
            pinfo: PrecursorInformation
            for msn_time, peaks, pinfo in msn_spectra:
                pinfo.precursor_scan_id = ms1_scan.id
                msn_scans.append(make_scan(
                    id=f"merged={self.scan_counter}",
                    ms_level=2, scan_time=msn_time,
                    index=self.scan_counter,
                    deconvoluted_peak_set=peaks,
                    precursor_information=pinfo))
                self.scan_counter += 1


            yield ScanBunch(ms1_scan, msn_scans)


PrecursorProductCorrelatingIterator.log_with_logger(logger)


class NaiveIonMobilityOverlapIterator(LogUtilsMixin):
    frame_pair_iterator: Union[Iterator[ScanBunch],
                               IonMobilitySourceRandomAccessFrameSource]

    def __init__(self, frame_pair_iterator):
        self.frame_pair_iterator = frame_pair_iterator
        self.scan_counter = 0

    def make_ms1_scan(self, precursor: IonMobilityFrame) -> WrappedScan:
        peaks_for_scan = [
            IonMobilityDeconvolutedPeak(
                f.neutral_mass, f.intensity, f.charge, f.intensity, 0,
                0, 0, 0, 0,
                score=f.score,
                envelope=f.sum_envelopes(),
                mz=f.mz,
                drift_time=f.apex_time)
            for f in
            precursor.deconvoluted_features
        ]

        ms1_scan = make_scan(
            id=precursor.id,
            ms_level=1, scan_time=precursor.time,
            index=self.scan_counter,
            deconvoluted_peak_set=DeconvolutedPeakSet(peaks_for_scan))

        self.scan_counter += 1
        return ms1_scan

    def __len__(self):
        return len(self.frame_pair_iterator)

    def split_products(self, ms1_scan: ScanBase, precursor: IonMobilityFrame, product: IonMobilityFrame) -> List[WrappedScan]:
        msn_scans = []
        for precursor_feature in precursor.deconvoluted_features:
            peaks = product.to_deconvoluted_peak_set(
                (precursor_feature.start_time,
                 precursor_feature.end_time),
                precursor_feature.neutral_mass)
            if not peaks:
                continue
            pinfo = PrecursorInformation(
                precursor_feature.mz,
                precursor_feature.intensity,
                precursor_feature.charge,
                ms1_scan.id,
                None,
                precursor_feature.neutral_mass,
                precursor_feature.charge,
                precursor_feature.intensity)
            pinfo._ion_mobility.add_ion_mobility(
                ion_mobility_drift_time,
                precursor_feature.apex_time
            )
            scan = make_scan(
                id=f"merged={self.scan_counter}",
                ms_level=2, scan_time=product.time,
                index=self.scan_counter,
                activation=ActivationInformation('cid', 2),
                deconvoluted_peak_set=peaks,
                precursor_information=pinfo)

            scan.annotations['source frame'] = product.id
            self.scan_counter += 1
            msn_scans.append(scan)
        return msn_scans

    def __iter__(self):
        return self

    def __next__(self):
        bunch: ScanBunch = next(self.frame_pair_iterator)
        ms1_scan = self.make_ms1_scan(bunch.precursor)
        msn_scans = []
        for product in bunch.products:
            msn_scans.extend(self.split_products(ms1_scan, bunch.precursor, product))
        return ScanBunch(ms1_scan, msn_scans)



"""This interface to the Bruker TDF file format is based upon the TimsData library
that is part of ProteoWizard (Apache II), and a common kernel of Python ctypes functions
found in dia-pasef and other Python wrappers of libtimsdata.
"""
import os
import re
import sqlite3
import sys
import warnings

from ctypes import cdll, c_char_p, c_uint32, c_int32, c_uint64, c_int64, c_double, c_void_p, POINTER, create_string_buffer
from weakref import WeakValueDictionary

import numpy as np

from ms_peak_picker import reprofile, pick_peaks, PeakIndex, PeakSet

from ms_deisotope.utils import Base
from ms_deisotope.data_source.metadata import software
from ms_deisotope.data_source.scan.loader import ScanDataSource, RandomAccessScanSource
from ms_deisotope.data_source.scan import PrecursorInformation, RawDataArrays
from ms_deisotope.data_source.scan.mobility_frame import IonMobilityFrame, IonMobilitySourceRandomAccessFrameSource

from ms_deisotope.data_source.metadata.activation import ActivationInformation, dissociation_methods
from ms_deisotope.data_source.metadata.scan_traits import (
    IsolationWindow, inverse_reduced_ion_mobility,
    ScanAcquisitionInformation, ScanWindow, ScanEventInformation)
from ms_deisotope.peak_dependency_network import Interval
from ms_deisotope.averagine import neutral_mass as calculate_neutral_mass

if sys.platform[:5] == "win32" or sys.platform[:5] == "win64":
    libname = "timsdata.dll"
elif sys.platform[:5] == "linux":
    libname = "libtimsdata.so"
else:
    raise Exception("Unsupported platform.")


dll = None


def load_library(search_paths=None):
    if search_paths is None:
        search_paths = []
    elif isinstance(search_paths, str):
        search_paths = [search_paths]
    global dll
    if dll is None:
        for lib_path in search_paths:
            try:
                dll = _load_library(lib_path)
            except (Exception) as err:
                continue
            if dll is not None:
                break
    return dll


def _load_library(lib_path):
    dll = cdll.LoadLibrary(os.path.realpath(lib_path))
    dll.tims_open.argtypes = [c_char_p, c_uint32]
    dll.tims_open.restype = c_uint64

    dll.tims_close.argtypes = [c_uint64]
    dll.tims_close.restype = None

    dll.tims_get_last_error_string.argtypes = [c_char_p, c_uint32]
    dll.tims_get_last_error_string.restype = c_uint32

    dll.tims_has_recalibrated_state.argtypes = [c_uint64]
    dll.tims_has_recalibrated_state.restype = c_uint32

    dll.tims_read_scans_v2.argtypes = [
        c_uint64, c_int64, c_uint32, c_uint32, c_void_p, c_uint32]
    dll.tims_read_scans_v2.restype = c_uint32

    # dll.tims_oneoverk0_to_ccs_for_mz.argtypes = [c_double, c_int32, c_double]
    # dll.tims_oneoverk0_to_ccs_for_mz.restype = c_double

    # dll.tims_ccs_to_oneoverk0_for_mz.argtypes = [c_double, c_int32, c_double]
    # dll.tims_ccs_to_oneoverk0_for_mz.restype = c_double


    convfunc_argtypes = [c_uint64, c_int64, POINTER(
        c_double), POINTER(c_double), c_uint32]

    for fn in [dll.tims_index_to_mz,
               dll.tims_mz_to_index,
               dll.tims_scannum_to_oneoverk0,
               dll.tims_oneoverk0_to_scannum,
               dll.tims_scannum_to_voltage,
               dll.tims_voltage_to_scannum]:
        fn.argtypes = convfunc_argtypes
        fn.restype = c_uint32
    return dll


def throw_tims_error(dll_handle):
    """Raise the last thronw timsdata error as a
    :exc:`RuntimeError`
    """

    size = dll_handle.tims_get_last_error_string(None, 0)
    buff = create_string_buffer(size)
    dll_handle.tims_get_last_error_string(buff, size)
    raise RuntimeError(buff.value)


def msms_type_to_ms_level(enum):
    if enum == 0:
        return 1
    else:
        return 2


msms_type_to_label = {
    0: "MS1 Scan",
    2: "MS2 Scan",
    8: "PASEF MS2 Scan",
    9: "DIA-PASEF",
}


msms_type_to_metadata_table = {
    2: "FrameMsMsInfo",
    8: "PasefFrameMsMsInfo",
    9: "DiaFrameMsMsWindows"
}


class TIMSMetadata(object):
    def _read_global_metadata(self):
        self._acquisition_parameters = acquisition_parameters = {}
        self._instrument_configuration = instrument_configuration = {}
        self._software = software_map = {
            "acquisition_software": {},
            "control_software": {},
        }
        q = self.connection.execute("SELECT Key, Value FROM GlobalMetadata;")
        for key, value in q:
            if key == "AcquistionSoftware":
                software_map['acquisition_software']['name'] = value
            elif key == "AcquisitionSoftwareVersion":
                software_map['acquisition_software']['version'] = value
            elif key == "InstrmentFamily":
                instrument_configuration['instrument_family'] = value
            elif key == "InstrmentRevision":
                instrument_configuration['instrument_revision'] = value
            elif key == "InstrumentSerialNumber":
                instrument_configuration['serial_number'] = value
            elif key == "AcquisitionDateTime":
                acquisition_parameters['acquisition_date'] = value
            elif key == "OperatorName":
                acquisition_parameters['operator_name'] = value
            elif key == "MzAcqRangeLower":
                acquisition_parameters['scan_window_lower'] = float(value)
            elif key == "MzAcqRangeUpper":
                acquisition_parameters['scan_window_upper'] = float(value)

    def _build_frame_index(self):
        self._frame_counts = {}
        q = self.connection.execute(
            "SELECT MsMsType, Count(*) FROM Frames GROUP BY MsMsType;")
        total = 0
        for scan_type_enum, count in q:
            self._frame_counts[msms_type_to_label[scan_type_enum]] = count
            total += count
        self._frame_counts['Total'] = total
        frame_id_scan_count = self.connection.execute("SELECT Id, NumScans, Time FROM Frames ORDER BY Id;")
        count_to_id = []
        time_to_id = []
        total = 0
        for f_c in frame_id_scan_count:
            count_to_id.append([total, f_c['Id']])
            total += f_c['NumScans']
            time_to_id.append([f_c['Time'], f_c['Id']])
        self._total_scans = total
        self._scan_count_to_frame_id = np.array(count_to_id)
        self._scan_time_to_frame_id = np.array(time_to_id)

    def _read_metadata(self):
        self._read_global_metadata()
        self._build_frame_index()

    def frame_count(self):
        return self._frame_counts['Total']

    def scan_count(self):
        return self._total_scans


class TIMSFrameInformation(Base):
    def __init__(self, source, id, accumulation_time, max_intensity, msms_type, mz_calibration, num_peaks, num_scans,
                 polarity, property_group, ramp_time, scan_mode, summed_intensities, t1, t2, time, tims_calibration,
                 tims_id, pasef_precursors=None):
        if pasef_precursors is None:
            pasef_precursors = []
        self.source = source
        self.id = id
        self.accumulation_time = accumulation_time
        self.max_intensity = max_intensity
        self.msms_type = msms_type
        self.mz_calibration = mz_calibration
        self.num_peaks = num_peaks
        self.num_scans = num_scans
        self.polarity = polarity
        self.property_group = property_group
        self.ramp_time = ramp_time
        self.scan_mode = scan_mode
        self.summed_intensities = summed_intensities
        self.t1 = t1
        self.t2 = t2
        self.time = time
        self.tims_calibration = tims_calibration
        self.tims_id = tims_id
        self.pasef_precursors = pasef_precursors

    def get_scan(self, start, end=None):
        if end is None:
            return self.source.get_scan_by_id("frame=%d scan=%d" % (self.id, start))
        return self.source.get_scan_by_id("frame=%d startScan=%d endScan=%d" % (self.id, start, end))

    @classmethod
    def from_query(cls, source, rowdict):
        frame = cls(
            source,
            rowdict['Id'],
            rowdict['AccumulationTime'],
            rowdict['MaxIntensity'],
            rowdict['MsMsType'],
            rowdict['MzCalibration'],
            rowdict['NumPeaks'],
            rowdict['NumScans'],
            rowdict['Polarity'],
            rowdict['PropertyGroup'],
            rowdict['RampTime'],
            rowdict['ScanMode'],
            rowdict['SummedIntensities'],
            rowdict['T1'],
            rowdict['T2'],
            rowdict['Time'],
            rowdict['TimsCalibration'],
            rowdict['TimsId'],
        )
        return frame


class PASEFPrecursorInformation(Base):
    def __init__(self, frame_id, start_scan, end_scan, isolation_mz, isolation_width, collision_energy, monoisotopic_mz,
                 charge, average_scan_number, intensity, parent):
        self.frame_id = frame_id
        self.start_scan = start_scan
        self.end_scan = end_scan
        self.isolation_mz = isolation_mz
        self.isolation_width = isolation_width
        self.collision_energy = collision_energy
        self.monoisotopic_mz = monoisotopic_mz
        self.charge = charge
        self.average_scan_number = average_scan_number
        self.intensity = intensity
        self.parent = parent

    @property
    def neutral_mass(self):
        return calculate_neutral_mass(self.monoisotopic_mz, self.charge)

    @classmethod
    def from_query(cls, rowdict):
        pasef_pinfo = cls(
            rowdict['Frame'],
            rowdict['ScanNumBegin'],
            rowdict['ScanNumEnd'],
            rowdict['IsolationMz'],
            rowdict['IsolationWidth'],
            rowdict['CollisionEnergy'],
            rowdict['MonoisotopicMz'],
            rowdict['Charge'],
            rowdict['ScanNumber'],
            rowdict['Intensity'],
            rowdict['Parent']
        )
        return pasef_pinfo


class TIMSSpectrumDataBase(Base):
    def __init__(self, frame, start_scan, end_scan=None):
        self.frame = frame
        self.start_scan = start_scan
        self._end_scan = end_scan

    @property
    def end_scan(self):
        if self._end_scan is None:
            return self.start_scan + 1
        else:
            return self._end_scan

    def is_combined(self):
        if self._end_scan is not None and self._end_scan - self.start_scan > 1:
            return True
        return False

    def make_id_string(self):
        if self._end_scan is None:
            return "frame=%d scan=%d" % (self.frame.id, self.start_scan + 1)
        else:
            return "frame=%d startScan=%d endScan=%d" % (self.frame.id, self.start_scan + 1, self.end_scan + 1)


class TIMSPASEFSpectrumData(TIMSSpectrumDataBase):
    def __init__(self, frame, start_scan, pasef_precursor, end_scan=None):
        super(TIMSPASEFSpectrumData, self).__init__(
            frame, start_scan, end_scan)
        self.pasef_precursor = pasef_precursor


default_scan_merging_parameters = {
    "fwhm": 0.04,
    "dx": 0.001
}


class TIMSAPI(object):
    initial_frame_buffer_size = 128

    def _convert_callback(self, frame_id, input_data, func):
        if isinstance(input_data, np.ndarray) and input_data.dtype == np.float64:
            # already supports buffer protocol, no extra copy
            in_array = input_data
        else:
            # convert data to appropriate float data buffer
            in_array = np.array(input_data, dtype=np.float64)
        cnt = len(in_array)
        out = np.empty(shape=cnt, dtype=np.float64)
        success = func(self.handle, frame_id,
                       in_array.ctypes.data_as(POINTER(c_double)),
                       out.ctypes.data_as(POINTER(c_double)),
                       cnt)
        if success == 0:
            throw_tims_error(self.dll)
        return out

    def index_to_mz(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_index_to_mz)

    def mz_to_index(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_mz_to_index)

    def scan_number_to_one_over_K0(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_scannum_to_oneoverk0)

    def one_over_K0_to_scan_number(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_oneoverk0_to_scannum)

    def scan_number_to_voltage(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_scannum_to_voltage)

    def voltage_to_scan_number(self, frame_id, mzs):
        return self._convert_callback(frame_id, mzs, self.dll.tims_voltage_to_scannum)
    # Output: list of tuples (indices, intensities)

    def read_scans(self, frame_id, scan_begin, scan_end):
        # buffer-growing loop
        while True:
            cnt = int(self.initial_frame_buffer_size)
            buf = np.empty(shape=cnt, dtype=np.uint32)
            buffer_size_in_bytes = 4 * cnt

            required_len = self.dll.tims_read_scans_v2(self.handle, frame_id, scan_begin, scan_end,
                                                       buf.ctypes.data_as(
                                                           POINTER(c_uint32)),
                                                       buffer_size_in_bytes)
            if required_len == 0:
                throw_tims_error(self.dll)

            if required_len > buffer_size_in_bytes:
                if required_len > 16777216:
                    # arbitrary limit for now...
                    raise RuntimeError("Maximum expected frame size exceeded.")
                self.initial_frame_buffer_size = required_len / 4 + 1  # grow buffer
            else:
                break

        result = []
        d = scan_end - scan_begin
        for i in range(scan_begin, scan_end):
            npeaks = buf[i-scan_begin]
            indices = buf[d:d + npeaks]
            d += npeaks
            intensities = buf[d:d + npeaks]
            d += npeaks
            result.append((indices, intensities))

        return result

    def read_spectrum(self, frame_id, scan_begin, scan_end):
        scans = self.read_scans(frame_id, scan_begin, scan_end)
        # Summarize on a grid
        allind = []
        allint = np.array([], dtype=float)
        for scan in scans:
            indices = np.array(scan[0])
            if len(indices) > 0:
                intens = scan[1]
                allind = np.concatenate((allind, indices))
                allint = np.concatenate((allint, intens))
        allmz = self.index_to_mz(frame_id, allind)
        return allmz, allint

    def read_spectrum_grid(self, frame_id, scan_begin, scan_end):
        scans = self.read_scans(frame_id, scan_begin, scan_end)
        # Summarize on a grid
        scan_numbers = []
        scan_mzs = []
        scan_intensities = []
        for i, scan in enumerate(scans, scan_begin):
            indices = np.array(scan[0])
            if len(indices) > 0:
                intens = scan[1]
                mzs = self.index_to_mz(frame_id, indices)
                scan_numbers.append(i)
                scan_mzs.append(mzs)
                scan_intensities.append(intens)
            else:
                scan_numbers.append(i)
                scan_mzs.append(np.array([], float))
                scan_intensities.append(np.array([], float))
        return scan_numbers, scan_mzs, scan_intensities

    def _decode_activation_method(self, frame):
        mode = frame.scan_mode
        if mode in (2, 8, 9):
            method = dissociation_methods["collision-induced dissociation"]
        elif mode in (3, 4, 5):
            method = dissociation_methods['in-source collision-induced dissociation']
        else:
            print("Unknown Scan Mode %d, Unknown Dissociation. Returning CID" % (mode, ))
            method = dissociation_methods["collision-induced dissociation"]
        return method

    def _query_frame_by_id_number(self, frame_id):
        cursor = self.connection.execute(
            "SELECT * FROM Frames WHERE Id={0};".format(frame_id))
        frame = TIMSFrameInformation.from_query(self, dict(cursor.fetchone()))
        # MS1
        if frame.msms_type == 0:
            pass
        # PASEF MS2
        elif frame.msms_type == 8:
            pasef_cursor = self.connection.execute(
                """SELECT Frame, ScanNumBegin, ScanNumEnd, IsolationMz, IsolationWidth, CollisionEnergy, MonoisotopicMz,
                          Charge, ScanNumber, Intensity, Parent FROM PasefFrameMsMsInfo f JOIN Precursors p on p.id=f.precursor
                          WHERE Frame = ?
                          ORDER BY ScanNumBegin
                          """, (frame_id, ))
            frame.pasef_precursors.extend(
                map(PASEFPrecursorInformation.from_query, pasef_cursor))
        else:
            warnings.warn("No support for MSMSType %r yet" %
                          (frame.msms_type, ))
        return frame


class BrukerTIMSScanDataSource(RandomAccessScanSource):
    _scan_merging_parameters = default_scan_merging_parameters.copy()

    def _is_profile(self, scan):
        if scan.is_combined():
            return True
        return False

    def _scan_time(self, scan):
        return scan.frame.time

    def _scan_id(self, scan):
        return scan.make_id_string()

    def _scan_title(self, scan):
        return self._scan_id(scan)

    def _ms_level(self, scan):
        if scan.frame.msms_type == 0:
            return 1
        return 2

    def _polarity(self, scan):
        if scan.frame.polarity == "+":
            return 1
        elif scan.frame.polarity == '-':
            return -1
        else:
            return None

    def _activation(self, scan):
        method = self._decode_activation_method(scan.frame)
        precursor = self._locate_pasef_precursor_for(scan)
        if precursor is not None:
            collision_energy = precursor.collision_energy
        return ActivationInformation(method, collision_energy)

    def _locate_pasef_precursor_for(self, scan):
        if scan.is_combined():
            # raise ValueError("Cannot determine precursor for combined spectra yet")
            query_interval = Interval(scan.start_scan, scan.end_scan)
            matches = []
            for precursor in scan.frame.pasef_precursors:
                if query_interval.overlaps(Interval(precursor.start_scan, precursor.end_scan)):
                    matches.append(precursor)
            n_matches = len(matches)
            if n_matches == 0:
                return None
            elif n_matches > 1:
                raise ValueError("Multiple precursors found for scan interval!")
            else:
                return matches[0]

        else:
            scan_number = scan.start_scan
            for precursor in scan.frame.pasef_precursors:
                if precursor.start_scan <= scan_number < precursor.end_scan:
                    return precursor
            return None

    def _precursor_information(self, scan):
        if scan.frame.msms_type == 8:
            precursor = self._locate_pasef_precursor_for(scan)
            if precursor is not None:
                mz = precursor.monoisotopic_mz
                intensity = precursor.intensity
                charge = precursor.charge
                parent_frame_id = precursor.parent

                if not scan.is_combined():
                    current_drift_time = self.scan_number_to_one_over_K0(scan.frame.id, [scan.start_scan])
                else:
                    current_drift_time = self.scan_number_to_one_over_K0(
                        scan.frame.id, [np.ceil(precursor.average_scan_number)])

                parent_frame = self.get_frame_by_id(parent_frame_id)._data
                precursor_scan_id = TIMSPASEFSpectrumData(parent_frame, precursor.start_scan, None, precursor.end_scan).make_id_string()
                product_scan_id = scan.make_id_string()
                pinfo = PrecursorInformation(
                    mz, intensity, charge, precursor_scan_id,
                    product_scan_id=product_scan_id, source=self, annotations={
                        inverse_reduced_ion_mobility: current_drift_time,
                    })
                return pinfo
        return None

    def _isolation_window(self, scan):
        if scan.frame.msms_type == 8:
            precursor = self._locate_pasef_precursor_for(scan)
            if precursor is not None:
                width = precursor.isolation_width / 2
                window = IsolationWindow(width, precursor.isolation_mz, width)
                return window
        return None

    def _scan_index(self, scan):
        cursor = self.connection.execute("SELECT sum(NumScans) FROM Frames WHERE Id < ?", (scan.frame.id, ))
        result = cursor.fetchone()[0]
        if result is None:
            result = 0
        return result + scan.start_scan

    def _acquisition_information(self, scan):
        idx = self._scan_index(scan) - \
            self._scan_count_to_frame_id[scan.frame.id - 1, 0]
        drift_time = self.scan_number_to_one_over_K0(scan.frame.id, [idx])[0]
        scan_time = self._scan_time(scan)
        evt = ScanEventInformation(scan_time, [
            ScanWindow(self._acquisition_parameters['scan_window_lower'],
                       self._acquisition_parameters['scan_window_upper'])
                       ])
        evt._ion_mobility.add_ion_mobility(inverse_reduced_ion_mobility, drift_time)
        return ScanAcquisitionInformation('none', [evt])

    def _get_centroids(self, scan):
        mzs, intensities = self.read_spectrum(
            scan.frame.id, scan.start_scan, scan.end_scan)
        sort_mask = np.argsort(mzs)
        mzs = mzs[sort_mask]
        intensities = intensities[sort_mask]
        centroids = pick_peaks(mzs, intensities, peak_mode="centroid")
        if centroids is None:
            centroids = PeakIndex(
                np.array([], float), np.array([], float), [])
        return centroids

    def _scan_arrays(self, scan):
        if scan.is_combined():
            mzs, intensities = self.read_spectrum(
                scan.frame.id, scan.start_scan, scan.end_scan)
            if len(mzs) == 0:
                return np.array([], dtype=float), np.array([], dtype=float)
            sort_mask = np.argsort(mzs)
            mzs = mzs[sort_mask]
            intensities = intensities[sort_mask]
            centroids = pick_peaks(mzs, intensities, peak_mode="centroid")
            if centroids is None:
                return np.array([], dtype=float), np.array([], dtype=float)
            mzs, intensities = reprofile(
                centroids, dx=self._scan_merging_parameters['dx'],
                override_fwhm=self._scan_merging_parameters['fwhm'])
            return mzs, intensities
        else:
            mzs, intensities = self.read_spectrum(scan.frame.id, scan.start_scan, scan.end_scan)
            return mzs, intensities


class BrukerTIMSFrameSource(IonMobilitySourceRandomAccessFrameSource):
    def _frame_id(self, data):
        return "frame=%d startScan=%d endScan=%d" % (data.id, 1, data.num_scans)

    def _frame_index(self, data):
        return data.id - 1

    def _frame_time(self, data):
        return data.time

    def _frame_ms_level(self, data):
        if data.msms_type == 0:
            return 1
        return 2

    def _frame_start_scan_index(self, data):
        start_i, frame_id = self._scan_count_to_frame_id[data.id - 1, :]
        assert frame_id == data.id
        return start_i

    def _frame_end_scan_index(self, data):
        start_i, frame_id = self._scan_count_to_frame_id[data.id - 1, :]
        assert frame_id == data.id
        return start_i + data.num_scans

    def _frame_precursor_information(self, data):
        pinfos = []
        for pasef in data.pasef_precursors:
            mz = pasef.monoisotopic_mz
            inten = pasef.intensity
            charge = pasef.charge
            parent_merged_id = 'frame=%d startScan=%d endScan=%d' % (pasef.parent, pasef.start_scan, pasef.end_scan)
            product_merged_id = 'frame=%d startScan=%d endScan=%d' % (pasef.frame_id, pasef.start_scan, pasef.end_scan)
            pinfos.append(PrecursorInformation(
                mz, inten, charge, parent_merged_id, self,
                product_scan_id=product_merged_id, annotations={
                    "parent_frame_index": pasef.parent - 1,
                    "start_scan": pasef.start_scan,
                    "end_scan": pasef.end_scan,
                }))
        if not pinfos:
            return None
        return pinfos

    def _frame_activation(self, data):
        activations = []
        method = self._decode_activation_method(data)
        for pasef in data.pasef_precursors:
            collision_energy = pasef.collision_energy
            activations.append(ActivationInformation(method, collision_energy))
        if not activations:
            return None
        return activations

    def _frame_isolation_window(self, data):
        isolations = []
        for pasef in data.pasef_precursors:
            width = pasef.isolation_width / 2.
            isolations.append(IsolationWindow(width, pasef.isolation_mz, width))
        if not isolations:
            return None
        return isolations

    def _frame_polarity(self, data):
        if data.polarity == "+":
            return 1
        elif data.polarity == '-':
            return -1
        else:
            warnings.warn("Unknown polarity %r" % (data.polarity, ))
            return 1

    def get_frame_by_time(self, frame_time):
        i = np.searchsorted(self._scan_time_to_frame_id[:, 0], frame_time) - 1
        _time, fi = self._scan_time_to_frame_id[i, :]
        return self.get_frame_by_id(fi)

    def get_frame_by_index(self, index):
        return self.get_frame_by_id(index + 1)

    def get_frame_by_id(self, frame_id):
        if isinstance(frame_id, str):
            match = multi_scan_id_parser.search(frame_id)
            if match is None:
                raise KeyError(frame_id)
            frame_id = int(match.group(1))
        if frame_id in self._frame_cache:
            return self._frame_cache[frame_id]
        frame = self._query_frame_by_id_number(frame_id)
        frame = self._make_frame(frame)
        self._frame_cache[frame_id] = frame
        return frame

    def _ms2_frames_for_parent_id(self, parent_frame_id):
        product_frame_ids = self.connection.execute("""
            SELECT Frame
            FROM PasefFrameMsMsInfo JOIN Precursors ON Precursors.id = PasefFrameMsMsInfo.precursor
            WHERE Parent = ?;""", (parent_frame_id, )).fetchall()
        product_frames = [self.get_frame_by_id(
            i) for i, in product_frame_ids]
        return product_frames

    def _validate_frame(self, data):
        return True

    def _make_frame(self, data):
        return IonMobilityFrame(data, self)

    def _cache_frame(self, frame):
        pass

    def _default_frame_iterator(self, start_index=None):
        if start_index is None:
            start_index = 0
        for i in range(start_index, self.frame_count()):
            yield self._query_frame_by_id_number(i + 1)

    def make_frame_iterator(self, iterator=None, grouped=False):
        from ms_deisotope.data_source.scan.scan_iterator import (
            _SingleScanIteratorImpl, _InterleavedGroupedScanIteratorImpl)
        if iterator is None:
            iterator = self._default_frame_iterator()
        if grouped:
            strategy = _InterleavedGroupedScanIteratorImpl(
                iterator, self._make_frame, self._validate_frame, self._cache_frame)
        else:
            strategy = _SingleScanIteratorImpl(
                iterator, self._make_frame, self._validate_frame, self._cache_frame)
        return strategy

    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        if scan_id is not None:
            frame = self.get_frame_by_id(scan_id)
            start_index = frame.index
        elif index is not None:
            start_index = index
        elif rt is not None:
            frame = self.get_frame_by_time(rt)
            start_index = frame.index
        if require_ms1:
            base_frame = self.get_frame_by_index(start_index)
            if base_frame.ms_level > 1:
                precursors = base_frame.precursor_information

            while start_index != 0:
                frame = self.get_frame_by_index(start_index)
                if frame.ms_level > 1:
                    start_index -= 1
                else:
                    break
        iterator = self.make_frame_iterator(
            self._default_frame_iterator(start_index), grouped=grouped)
        return iterator


single_scan_id_parser = re.compile(r"frame=(\d+) scan=(\d+)")
multi_scan_id_parser = re.compile(r"frame=(\d+) startScan=(\d+) endScan=(\d+)")


class BrukerTIMSLoader(BrukerTIMSFrameSource, BrukerTIMSScanDataSource, TIMSMetadata, TIMSAPI):

    def __init__(self, analysis_directory, use_recalibrated_state=False, scan_merging_parameters=None):
        if sys.version_info.major == 2:
            if isinstance(analysis_directory, str):
                analysis_directory = analysis_directory.decode('utf8')
            if not isinstance(analysis_directory, unicode):
                raise ValueError("analysis_directory must be a Unicode string.")
        if sys.version_info.major == 3:
            if isinstance(analysis_directory, bytes):
                analysis_directory = analysis_directory.decode('utf8')
            if not isinstance(analysis_directory, str):
                raise ValueError("analysis_directory must be a string.")
        if scan_merging_parameters is None:
            scan_merging_parameters = default_scan_merging_parameters.copy()
        else:
            for key, value in default_scan_merging_parameters.items():
                scan_merging_parameters.setdefault(key, value)

        self.dll = load_library()

        self.handle = self.dll.tims_open(
            analysis_directory.encode('utf-8'), 1 if use_recalibrated_state else 0)
        if self.handle == 0:
            throw_tims_error(self.dll)

        self._source_file_name = analysis_directory
        self.connection = sqlite3.connect(os.path.join(analysis_directory, "analysis.tdf"))
        self.connection.row_factory = sqlite3.Row

        self.initialize_frame_cache()
        self.initialize_scan_cache()

        self.initial_frame_buffer_size = 128 # may grow in readScans()
        self._read_metadata()
        self._scan_merging_parameters = scan_merging_parameters
        self._producer = self.make_frame_iterator(grouped=True)

    @property
    def source_file_name(self):
        return self._source_file_name

    def __del__(self):
        if hasattr(self, 'handle'):
            self.dll.tims_close(self.handle)

    def _describe_frame(self, frame_id):
        cursor = self.connection.execute("SELECT * FROM Frames WHERE Id={0};".format(frame_id))
        return dict(cursor.fetchone())

    def __len__(self):
        return self.frame_count()

    def next(self):
        return next(self._producer)

    def get_scan_by_index(self, scan_index):
        frame_index = np.searchsorted(
            self._scan_count_to_frame_id[:, 0], scan_index + 1)
        if frame_index > 0:
            frame_index -= 1
        c, i = self._scan_count_to_frame_id[frame_index, :]
        frame_id = i
        remainder = scan_index - c + 1
        scan_number = remainder
        scan_id = "frame=%d scan=%d" % (frame_id, scan_number)
        return self.get_scan_by_id(scan_id)

    def get_scan_by_time(self, scan_time):
        i = np.searchsorted(self._scan_time_to_frame_id[:, 0], scan_time) - 1
        time, fi = self._scan_time_to_frame_id[i, :]
        remainder = scan_time - time
        try:
            next_time = self._scan_time_to_frame_id[i + 1, 0]
        except IndexError:
            # An instance with exactly one frame?
            if i == 0:
                next_time = time * 2.0
            else:
                raise
        duration = next_time - time
        n_scans = self.get_frame_by_id(fi)._data.num_scans
        grid_space = np.linspace(0, duration, n_scans)
        i = np.searchsorted(grid_space, remainder) - 1
        scan_id = 'frame=%d scan=%d' % (fi, i)
        return self.get_scan_by_id(scan_id)

    def get_scan_by_id(self, scan_id):
        match = single_scan_id_parser.match(scan_id)
        if match is None:
            match = multi_scan_id_parser.match(scan_id)
            if match is None:
                raise ValueError("%r does not look like a TIMS nativeID" % (scan_id, ))
            else:
                frame_id, start_scan, end_scan = map(int, match.groups())
                frame = self.get_frame_by_index(frame_id - 1)._data
                if frame.msms_type == 8:
                    pasef_scan = TIMSPASEFSpectrumData(
                        frame, start_scan - 1, None, end_scan)
                    pasef_scan.pasef_precursor = self._locate_pasef_precursor_for(pasef_scan)
                    scan_obj = self._make_scan(pasef_scan)
                else:
                    scan_obj = self._make_scan(TIMSSpectrumDataBase(frame, start_scan - 1, end_scan))
        else:
            frame_id, scan_number = map(int, match.groups())
            frame = self.get_frame_by_index(frame_id - 1)._data
            if frame.msms_type == 8:
                pasef_scan = TIMSPASEFSpectrumData(frame, scan_number - 1, None)
                pasef_scan.pasef_precursor = self._locate_pasef_precursor_for(pasef_scan)
                scan_obj = self._make_scan(pasef_scan)
            else:
                scan_obj = self._make_scan(TIMSSpectrumDataBase(frame, scan_number - 1))
        # Cache scan here
        return scan_obj

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        """Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        This method will trigger several random-access operations, making it prohibitively
        expensive for normally compressed files.

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        """
        raise NotImplementedError()

    def _ms1_frame_iterator(self):
        ms1_frame_ids = self.connection.execute("SELECT Id FROM Frames WHERE MsMsType = 0;").fetchall()
        for frame_id, in ms1_frame_ids:
            yield frame_id

    def _make_default_iterator(self, frame=False):
        if frame:
            return self._ms1_frame_iterator()
        else:
            raise NotImplementedError()



# cython: embedsignature=True
#

from libc.stdlib cimport malloc, free

cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x)

cimport cython

from cpython cimport PyErr_SetString
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append, PyList_GetItem, PyList_SetItem, PyList_New
from ms_peak_picker._c.peak_set cimport FittedPeak
from brainpy._c.isotopic_distribution cimport TheoreticalPeak, _isotopic_variants
from brainpy._c.double_vector cimport (
    DoubleVector as dvec,
    make_double_vector,
    free_double_vector,
    double_vector_append,
    double_vector_to_list)

from ms_deisotope._c.scoring cimport IsotopicFitterBase
from ms_deisotope._c.averagine cimport (
    AveragineCache, PROTON, neutral_mass as calc_neutral_mass,
    TheoreticalIsotopicPattern)

from ms_deisotope._c.feature_map.feature_map cimport LCMSFeatureMap
from ms_deisotope._c.feature_map.feature_fit cimport LCMSFeatureSetFit
from ms_deisotope._c.feature_map.lcms_feature cimport (
    LCMSFeature,
    FeatureSetIterator,
    EmptyFeature)

cimport numpy as np
import numpy as np


np.import_array()


cdef object zeros = np.zeros


cdef double INF = float('inf')


cdef object double_vector_to_ndarray(dvec* vec):
    cdef:
        np.ndarray[double, ndim=1, mode='c'] out
        np.npy_intp shape[1]
        size_t i, n

    n = shape[0] = vec.used
    # out = np.PyArray_SimpleNew(1, shape, np.NPY_DOUBLE)
    out = zeros(n)
    for i in range(n):
        out[i] = vec.v[i]
    # The NDArray returned here seems to get corrupted somehow.
    # Converting it to a managed object here and converting it back
    # into a NumPy array later at smooth_leveled seems to prevent it.
    return tuple(out)


@cython.final
cdef class CartesianProductIterator(object):
    cdef:
        public list collections
        public size_t size
        int* lengths
        int* indices
        public bint done
        public size_t total_combinations

    cdef inline int _initialize(self, list collections) except 1:
        cdef:
            size_t i, n
            list sublist

        self.collections = collections
        self.size = PyList_GET_SIZE(self.collections)
        self.lengths = <int*>malloc(sizeof(int) * self.size)
        if self.lengths == NULL:
            PyErr_SetString(MemoryError, "Failed to allocate length array in CartesianProductIterator")
        self.indices = <int*>malloc(sizeof(int) * self.size)
        if self.indices == NULL:
            PyErr_SetString(MemoryError, "Failed to allocate indices array in CartesianProductIterator")
        self.done = False
        self.total_combinations = 1

        for i in range(self.size):
            sublist = <list>PyList_GetItem(self.collections, i)
            self.lengths[i] = PyList_GET_SIZE(sublist)
            self.total_combinations *= self.lengths[i]
            if self.lengths[i] == 0:
                self.done = True
            self.indices[i] = 0

    @staticmethod
    cdef CartesianProductIterator _create(list collections):
        cdef CartesianProductIterator self = CartesianProductIterator.__new__(CartesianProductIterator)
        self._initialize(collections)
        return self

    def __init__(self, collections):
        self.collections = list(map(list, collections))
        self._initialize(self.collections)

    def __dealloc__(self):
        free(self.lengths)
        free(self.indices)

    def get_lengths(self):
        return [self.lengths[i] for i in range(self.size)]

    def get_indices(self):
        return [self.indices[i] for i in range(self.size)]

    cdef inline bint has_more(self):
        return not self.done

    cdef inline list compose_next_value(self):
        cdef:
            int i, ix
            list result, sublist
            object value
        if self.done:
            return None
        result = []
        for i in range(self.size):
            sublist = <list>PyList_GetItem(self.collections, i)
            ix = self.indices[i]
            value = <object>PyList_GetItem(sublist, ix)
            result.append(value)
        return result

    cdef inline void advance(self):
        for i in range(self.size - 1, -1, -1):
            if self.indices[i] == self.lengths[i] - 1:
                self.indices[i] = 0
                if i == 0:
                    self.done = True
            else:
                self.indices[i] += 1
                break

    cdef inline list get_next_value(self):
        cdef list value
        if self.done:
            return None
        else:
            value = self.compose_next_value()
            self.advance()
            return value

    def __iter__(self):
        return self

    def __next__(self):
        if self.done:
            raise StopIteration()
        value = self.compose_next_value()
        if value is None:
            raise StopIteration()
        else:
            self.advance()
        return value


cdef tuple _conform_envelopes(list experimental, list base_theoretical, size_t* n_missing_out,
                              double minimum_theoretical_abundance=0.05):
    cdef:
        envelope_conformer conformer
        list cleaned_eid, tid

    conformer = envelope_conformer._create()
    conformer.acquire(experimental, base_theoretical)
    conformer.conform(minimum_theoretical_abundance)

    cleaned_eid = conformer.experimental
    tid = conformer.theoretical
    n_missing_out[0] = conformer.n_missing
    return (cleaned_eid, tid)


@cython.final
cdef class envelope_conformer:
    cdef:
        list experimental
        list theoretical
        size_t n_missing

    @staticmethod
    cdef envelope_conformer _create():
        return envelope_conformer.__new__(envelope_conformer)

    cdef inline void acquire(self, list experimental, list theoretical):
        self.experimental = experimental
        self.theoretical = theoretical
        self.n_missing = 0

    cdef inline void conform(self, double minimum_theoretical_abundance=0.05):
        cdef:
            double total = 0
            size_t n_missing = 0
            size_t i = 0
            list cleaned_eid
            list tid
            FittedPeak fpeak
            TheoreticalPeak tpeak, peak

        cleaned_eid = []

        for i in range(PyList_GET_SIZE(self.experimental)):
            fpeak = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
            if fpeak is None:
                peak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
                fpeak = FittedPeak._create(peak.mz, 1, 1, -1, -1, 0, 1, 0, 0)
                if peak.intensity > minimum_theoretical_abundance:
                    n_missing += 1
            total += fpeak.intensity
            cleaned_eid.append(fpeak)
            i += 1

        tid = []
        for i in range(PyList_GET_SIZE(self.theoretical)):
            tpeak = <TheoreticalPeak>PyList_GET_ITEM(self.theoretical, i)
            peak = tpeak.clone()
            peak.intensity *= total
            tid.append(peak)
        self.experimental = cleaned_eid
        self.theoretical = tid
        self.n_missing = n_missing


cpdef tuple conform_envelopes(list experimental, list base_theoretical,
                              double minimum_theoretical_abundance=0.01):
    cdef:
        size_t n_missing
        tuple out

    out = _conform_envelopes(experimental, base_theoretical, &n_missing, minimum_theoretical_abundance)
    out += (n_missing,)
    return out


cdef int has_no_valid_features(list features):
    cdef:
        size_t i, n, cnt
        object feature

    n = PyList_GET_SIZE(features)
    cnt = 0
    for i in range(n):
        feature = <object>PyList_GET_ITEM(features, i)
        if feature is None:
            cnt += 1
    return cnt == n


cdef size_t MAX_COMBINATIONS = 10000 // 2


cdef class LCMSFeatureProcessorBase(object):
    cdef:
        public LCMSFeatureMap feature_map
        public IsotopicFitterBase scorer
        public AveragineCache averagine

    cpdef TheoreticalIsotopicPattern create_theoretical_distribution(self, double mz, int charge, double charge_carrier=PROTON, double truncate_after=0.8,
                                                                     double ignore_below=0.05):
        cdef:
            TheoreticalIsotopicPattern base_tid
        base_tid = self.averagine.isotopic_cluster(
            mz, charge, truncate_after=truncate_after, ignore_below=ignore_below,
            charge_carrier=charge_carrier)
        return base_tid

    cpdef list find_all_features(self, double mz, double error_tolerance=2e-5):
        return self.feature_map.find_all(mz, error_tolerance)

    cpdef list find_features(self, double mz, double error_tolerance=2e-5, LCMSFeature interval=None):
        cdef:
            list f, overlapped
            double search_width, hit_width
            LCMSFeature el
            size_t i, n
            bint has_miss
        has_miss = False
        f = self.find_all_features(mz, error_tolerance)
        if not f:
            f = [None]
            has_miss = True
        elif interval is not None:
            overlapped = []
            n = PyList_GET_SIZE(f)
            search_width = interval.get_end_time() - interval.get_start_time()
            if search_width == 0:
                return [None]
            for i in range(n):
                el = <LCMSFeature>PyList_GET_ITEM(f, i)
                if el.overlaps_in_time(interval):
                    hit_width = el.get_end_time() - el.get_start_time()
                    if (hit_width / search_width) < 0.05:
                        continue
                    overlapped.append(el)
            f = overlapped
            if PyList_GET_SIZE(f) == 0:
                f = [None]
        return f

    cpdef list match_theoretical_isotopic_distribution(self, list theoretical_distribution, double error_tolerance=2e-5, LCMSFeature interval=None):
        """Given a list of theoretical peaks, find their counterparts in :attr:`peaklist` within `error_tolerance`
        ppm error. If no experimental peak is found, a placeholder will be used in its stead.

        Parameters
        ----------
        theoretical_distribution : list of TheoreticalPeak
            The theoretical isotopic pattern to match
        error_tolerance : float, optional
            Parts-per-million error tolerance to permit in searching for matches

        Returns
        -------
        list of list of LCMSFeature
            The list of matched features for each theoretical m/z
        """
        cdef:
            list experimental_distribution, found
            size_t i, n
            TheoreticalPeak p

        experimental_distribution = []
        n = PyList_GET_SIZE(theoretical_distribution)
        for i in range(n):
            p = <TheoreticalPeak>PyList_GET_ITEM(theoretical_distribution, i)
            found = self.find_features(p.mz, error_tolerance, interval=interval)
            experimental_distribution.append(found)
        return experimental_distribution

    cpdef tuple conform_envelopes(self, list experimental, list base_theoretical):
        return conform_envelopes(experimental, base_theoretical)

    cdef tuple _conform_envelopes(self, list experimental, list base_theoretical, size_t* n_missing):
        return _conform_envelopes(experimental, base_theoretical, n_missing)

    cdef double _find_thresholded_score(self, dvec* scores, double percentage):
        cdef:
            size_t i, n, count
            double maximum, val, threshold, acc

        maximum = -INF
        count = 0
        acc = 0
        n = scores.used
        for i in range(n):
            val = scores.v[i]
            if val > maximum:
                maximum = val
        threshold = maximum * percentage
        for i in range(n):
            val = scores.v[i]
            if val > threshold:
                acc += val
                count += 1
        if count == 0:
            return 0
        if count == 1:
            # Penalize singletons? This may harm more than help
            count += 1
        return acc / count

    cpdef LCMSFeatureSetFit _fit_single_feature_set(self, list features, list base_tid, double error_tolerance,
                                                    int charge, double charge_carrier=PROTON, int max_missed_peaks=1,
                                                    double threshold_scale=0.3):
        cdef:
            double score, final_score, score_acc, neutral_mass
            envelope_conformer conformer
            dvec* score_vec
            dvec* time_vec
            np.ndarray scores_array, times_array
            FeatureSetIterator feat_iter
            size_t feat_i, feat_n, n_missing, missing_features, counter

        conformer = envelope_conformer._create()
        feat_iter = FeatureSetIterator._create(features)
        score_acc = 0
        counter = 0
        score_vec = make_double_vector()
        time_vec = make_double_vector()
        while feat_iter.has_more():
            eid = feat_iter.get_next_value()
            if eid is None:
                continue
            conformer.acquire(eid, base_tid)
            conformer.conform()
            cleaned_eid = conformer.experimental
            tid = conformer.theoretical
            n_missing = conformer.n_missing
            if n_missing > max_missed_peaks:
                continue
            score = self.scorer._evaluate(None, cleaned_eid, tid)
            if npy_isnan(score):
                continue
            score_acc += score
            counter += 1
            double_vector_append(score_vec, score)
            double_vector_append(time_vec, feat_iter.get_current_time())
        # Compute feature score from score_vec
        final_score = self._find_thresholded_score(score_vec, threshold_scale)
        missing_features = 0
        feat_n = PyList_GET_SIZE(features)
        for feat_i in range(feat_n):
            f = <LCMSFeature?>PyList_GET_ITEM(features, feat_i)
            if f is None:
                missing_features += 1
        f = <LCMSFeature?>PyList_GET_ITEM(features, 0)
        neutral_mass = calc_neutral_mass(
                f.get_mz(), charge, charge_carrier)
        scores_array = double_vector_to_ndarray(score_vec)
        times_array = double_vector_to_ndarray(time_vec)
        fit = LCMSFeatureSetFit._create(
            features, base_tid, final_score, charge, missing_features,
            [], None, neutral_mass, counter, scores_array, times_array)
        free_double_vector(score_vec)
        free_double_vector(time_vec)
        return fit

    cdef list _fit_theoretical_distribution_on_features(self, double mz, double error_tolerance, int charge,
                                                        TheoreticalIsotopicPattern base_tid,
                                                        double charge_carrier=PROTON,
                                                        double truncate_after=0.8, int max_missed_peaks=1,
                                                        double threshold_scale=0.3, LCMSFeature feature=None):
        cdef:
            double score, final_score, score_acc, neutral_mass
            list feature_groups, feature_fits, features
            list eid, cleaned_eid, tid
            object scores_array, times_array
            tuple temp
            size_t combn_size, combn_i, n_missing, missing_features, counter
            TheoreticalIsotopicPattern snapped_tid
            size_t feat_i, feat_n
            FeatureSetIterator feat_iter
            LCMSFeature f
            CartesianProductIterator combn_iter
            envelope_conformer conformer
            dvec* score_vec
            dvec* time_vec

        feature_groups = self.match_theoretical_isotopic_distribution(
            base_tid.peaklist, error_tolerance, interval=feature)
        feature_fits = []

        conformer = envelope_conformer._create()

        combn_iter = CartesianProductIterator._create(feature_groups)
        combn_i = 0
        while combn_iter.has_more():
            features = combn_iter.get_next_value()
            if features is None:
                break
            combn_i += 1
            if combn_i > MAX_COMBINATIONS:
                break
            elif combn_i % (100000) == 0:
                print("\t%d combinations have been considered for m/z %0.3f at charge %d in %s (%0.2f%%)" % (
                    combn_i, mz, charge, "-" if feature is None else (
                        "%0.2f-%0.2f" % (feature.start_time, feature.end_time)),
                    100. * combn_i / combn_iter.total_combinations))

            if has_no_valid_features(features):
                continue
            # If the monoisotopic feature wasn't actually observed, create a dummy feature
            # since the monoisotopic feature cannot be None
            # print("Checking monoisotopic feature")
            if features[0] is None:
                # print("Creating monoisotopic feature stub")
                features = list(features)
                features[0] = EmptyFeature._create(mz)
                snapped_tid = base_tid.clone().shift(mz)
            else:
                f = <LCMSFeature>PyList_GET_ITEM(features, 0)
                snapped_tid = base_tid.clone().shift(f.get_mz())
            # feat_iter = FeatureSetIterator._create_with_threshold(features, base_tid, 0.1)
            feat_iter = FeatureSetIterator._create(features)
            counter = 0
            score_acc = 0

            score_vec = make_double_vector()
            time_vec = make_double_vector()
            # print("Scoring features")
            while feat_iter.has_more():
                eid = feat_iter.get_next_value()
                if eid is None:
                    continue
                counter += 1
                conformer.acquire(eid, snapped_tid.peaklist)
                conformer.conform()
                cleaned_eid = conformer.experimental
                tid = conformer.theoretical
                n_missing = conformer.n_missing
                if n_missing > max_missed_peaks:
                    continue

                score = self.scorer._evaluate(None, cleaned_eid, tid)
                if npy_isnan(score):
                    continue
                score_acc += score
                double_vector_append(score_vec, score)
                double_vector_append(time_vec, feat_iter.get_current_time())
            if score_vec.used == 0:
                free_double_vector(score_vec)
                free_double_vector(time_vec)
                continue
            # print(f"Scored over {counter} points")
            # Compute feature score from score_vec
            final_score = self._find_thresholded_score(
                score_vec, threshold_scale)
            # print(f"Final Score {final_score}")
            missing_features = 0
            feat_n = PyList_GET_SIZE(features)
            for feat_i in range(feat_n):
                f = <LCMSFeature>PyList_GET_ITEM(features, feat_i)
                if f is None:
                    missing_features += 1
            f = <LCMSFeature>PyList_GET_ITEM(features, 0)
            neutral_mass = calc_neutral_mass(
                    snapped_tid.get_monoisotopic_mz(), charge, charge_carrier)
            # print("Preparing to hand off arrays")
            scores_array = double_vector_to_ndarray(score_vec)
            times_array = double_vector_to_ndarray(time_vec)
            # print("Creating fit")

            fit = LCMSFeatureSetFit._create(
                features, snapped_tid,
                final_score, charge,
                missing_features, [],
                None, neutral_mass,
                counter, scores_array,
                times_array)

            # print("Freeing arrays")
            free_double_vector(score_vec)
            free_double_vector(time_vec)
            if self.scorer.reject_score(fit.score):
                continue
            feature_fits.append(fit)
        # print(f"Completed generating fits for {mz} @ {charge}: {combn_i}")
        return feature_fits

    cpdef list _fit_feature_set(self, double mz, double error_tolerance, int charge, int left_search=1,
                                int right_search=1, double charge_carrier=PROTON, double truncate_after=0.8,
                                int max_missed_peaks=1, double threshold_scale=0.3, LCMSFeature feature=None):
        cdef:
            list feature_fits
            TheoreticalIsotopicPattern base_tid
        base_tid = self.create_theoretical_distribution(
            mz, charge, charge_carrier, truncate_after)
        # print(f"Fitting feature set for {mz} at charge {charge}")
        feature_fits = self._fit_theoretical_distribution_on_features(
            mz, error_tolerance, charge, base_tid,
            charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
            feature)
        # print(f"Fitted {len(feature_fits)}")
        return feature_fits

    cpdef list _fit_composition(self, dict composition, double error_tolerance, int charge, double charge_carrier=PROTON,
                                double truncate_after=0.8, double ignore_below=0.05, int max_missed_peaks=1,
                                double threshold_scale=0.3, LCMSFeature feature=None):
        cdef:
            list base_tid, kept_tid, feature_fits
            double total, cumsum, mz
            TheoreticalPeak tp
            size_t i, n

        # Can't refactor shared components with `create_theoretical_isotopic_distribution` because
        # this needs to include its own truncate-after component as well as ignore-under
        base_tid = _isotopic_variants(composition, npeaks=None, charge=charge, charge_carrier=charge_carrier)
        n = PyList_GET_SIZE(base_tid)
        total = 0
        cumsum = 0
        kept_tid = []
        for i in range(n):
            tp = <TheoreticalPeak>PyList_GET_ITEM(base_tid, i)
            cumsum += tp.intensity
            if (tp.intensity < ignore_below) and (i > 1):
                pass
            else:
                PyList_Append(kept_tid, tp)
                total += tp.intensity
            if cumsum >= truncate_after:
                break

        n = PyList_GET_SIZE(kept_tid)
        for i in range(n):
            tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, i)
            tp.intensity /= total

        if n == 0:
            return []

        tp = <TheoreticalPeak>PyList_GET_ITEM(kept_tid, 0)
        mz = tp.mz

        feature_fits = self._fit_theoretical_distribution_on_features(
            mz, error_tolerance, charge, kept_tid,
            charge_carrier, truncate_after, max_missed_peaks, threshold_scale,
            feature)
        return feature_fits



"""A collection of utilities for using :mod:`ms_deisotope` to convert between mass spectrometry
data file formats.
"""
import os

import click
from six import string_types as basestring

from ms_peak_picker.scan_filter import parse as parse_filter
from pyteomics.xml import unitfloat

import ms_deisotope
from ms_deisotope.data_source._compression import GzipFile
from ms_deisotope.data_source._buffer import PreBufferedStreamReader
from ms_deisotope.data_source.metadata import activation as activation_module, data_transformation
from ms_deisotope.output import MzMLSerializer, MGFSerializer

from ms_deisotope.tools.utils import is_debug_mode, register_debug_hook, progress, type_cast


try:
    from ms_deisotope.output import MzMLbSerializer
except ImportError:
    MzMLbSerializer = None


@click.group()
def ms_conversion():
    """A command line tool for converting between mass spectrometry data formats
    """


def to_mgf(reader, outstream, msn_filters=None):
    """Translate the spectra from `reader` into MGF format written to `outstream`.

    As MGF files do not usually contain MS1 spectra, these will be omitted. Additionally,
    MSn spectra will be centroided if they are not already.

    Parameters
    ----------
    reader : :class:`~.ScanIterator`
        The source of spectra to iterate over
    outstream : file-like
        The output stream to write to
    msn_filters : list, optional
        An optional list of strings or :class:`~.ScanFilterBase` instances which will be
        used to transform the m/z and intennsity arrays of MSn spectra before they are futher
        procssed. (the default is None, which results in no transformations)

    """
    if not msn_filters:
        msn_filters = []
    reader.make_iterator(grouped=False)
    writer = MGFSerializer(outstream, deconvoluted=False)
    try:
        n_spectra = len(reader)
    except TypeError:
        n_spectra = None
    progbar = progress(
        reader,
        label="Processed Spectra", length=n_spectra,
        item_show_func=lambda x: str(x.id) if x else '',
        color=True, fill_char=click.style('-', 'green'))
    with progbar:
        for scan in progbar:
            if scan.ms_level == 1:
                continue
            if msn_filters:
                scan = scan.transform(msn_filters)
            if scan.peak_set is None:
                scan.pick_peaks()
            writer.save_scan(scan)
    outstream.flush()


@ms_conversion.command('mgf', short_help="Convert a mass spectrometry data file to MGF")
@click.argument("source")
@click.argument("output", type=click.Path(writable=True))
@click.option("-z", "--compress", is_flag=True, help=("Compress the output file using gzip"))
@click.option("-rn", "--msn-filter", "msn_filters", multiple=True, type=parse_filter)
def mgf(source, output, compress=False, msn_filters=None):
    """Convert a mass spectrometry data file to MGF. MGF can only represent centroid spectra
    and generally does not contain any MS1 information.
    """
    if compress:
        if not output.endswith(".gz") and output != '-':
            output += '.gz'
        stream = click.open_file(output, 'wb')
        stream = GzipFile(fileobj=stream, mode='wb')
    else:
        stream = click.open_file(output, 'wb')
    use_index = True
    if source == "-":
        click.secho("Reading input file from STDIN, some file formats will not be supported.", err=True, fg='yellow')
        source = PreBufferedStreamReader(click.open_file(source, mode='rb'))
        # Cannot use the offset index of a file we cannot seek through
        use_index = False

    reader = ms_deisotope.MSFileLoader(source, use_index=use_index)
    to_mgf(reader, stream, msn_filters=msn_filters)


def to_mzml(reader, outstream, pick_peaks=False, reprofile=False, ms1_filters=None, msn_filters=None,
            default_activation=None, correct_precursor_mz=False, write_index=True, update_metadata=True,
            writer_type=None, compression='zlib', close=True, **kwargs):
    """Translate the spectra from `reader` into mzML format written to `outstream`.

    Wraps the process of iterating over `reader`, performing a set of simple data transformations if desired,
    and then converts each :class:`~.Scan` into mzML format. Any data transformations are described in the
    appropriate metadata section. All other metadata from `reader` is copied to into `outstream`.

    Parameters
    ----------
    reader : :class:`~.ScanIterator`
        The source of spectra to iterate over
    outstream : file-like
        The output stream to write mzML to.
    pick_peaks : bool, optional
        Whether to centroid profile spectra (the default is False)
    reprofile: bool, optional
        Whether to reprofile spectra from their centroids (the default is False)
    ms1_filters : list, optional
        An optional list of strings or :class:`~.ScanFilterBase` instances which will be
        used to transform the m/z and intensity arrays of MS1 spectra before they are further
        processed (the default is None, which results in no transformations)
    msn_filters : list, optional
        An optional list of strings or :class:`~.ScanFilterBase` instances which will be
        used to transform the m/z and intennsity arrays of MSn spectra before they are futher
        procssed. (the default is None, which results in no transformations)
    default_activation : :class:`str` or :class:`dict`, optional
        A default activation type to use when `reader` does not contain that information (the default is None)
    correct_precursor_mz : bool, optional
        Whether or not to assign the precursor m/z of each product scan to the nearest peak
        m/z in the precursor's peak list. (the default is False, which results in no correction)

    """
    if writer_type is None:
        writer_type = MzMLSerializer
    if ms1_filters is None:
        ms1_filters = []
    if msn_filters is None:
        msn_filters = []
    reader.make_iterator(grouped=True)
    writer = writer_type(
        outstream,
        len(reader),
        deconvoluted=False,
        build_extra_index=write_index,
        include_software_entry=update_metadata,
        compression=compression,
        close=close
    )
    writer.copy_metadata_from(reader)
    if update_metadata:
        method = data_transformation.ProcessingMethod(software_id='ms_deisotope_1')
        if pick_peaks:
            method.add('MS:1000035')
        if correct_precursor_mz:
            method.add('MS:1000780')
        if reprofile:
            method.add('MS:1000784')
        method.add('MS:1000544')
        writer.add_data_processing(method)
    if default_activation is not None:
        if isinstance(default_activation, basestring):
            default_activation = activation_module.ActivationInformation(
                default_activation, unitfloat(0, 'electronvolt'))
        elif isinstance(default_activation, dict):
            default_activation = activation_module.ActivationInformation(**default_activation)
        else:
            click.secho(f"Could not convert {default_activation} into ActivationInformation", err=1,
                        fg='yellow')
            default_activation = None
    if pick_peaks:
        try:
            writer.remove_file_contents("profile spectrum")
        except KeyError:
            pass
        writer.add_file_contents("centroid spectrum")
    n_spectra = len(reader)
    progbar = progress(
        label="Processed Spectra", length=n_spectra,
        item_show_func=lambda x: str(x.precursor.id if x.precursor else
                                     x.products[0].id) if x else '',
        color=True, fill_char=click.style('-', 'green'))
    with progbar:
        for bunch in reader:
            progbar.current_item = bunch
            progbar.update((bunch.precursor is not None) + len(bunch.products))
            discard_peaks = False
            if bunch.precursor is not None:
                if (reprofile):
                    bunch = bunch._replace(precursor=bunch.precursor.reprofile())
                if ms1_filters:
                    bunch = bunch._replace(precursor=bunch.precursor.transform(ms1_filters))
                if (pick_peaks or not bunch.precursor.is_profile):
                    bunch.precursor.pick_peaks()
                if correct_precursor_mz:
                    if not pick_peaks:
                        bunch.precursor.pick_peaks()
                        if bunch.precursor.is_profile:
                            discard_peaks = True

            for i, product in enumerate(bunch.products):
                # if reprofile:
                #     product = bunch.products[i] = product.reprofile()
                if msn_filters:
                    product = bunch.products[i] = product.transform(msn_filters)
                if pick_peaks or not product.is_profile:
                    product.pick_peaks()
                if product.activation is None and default_activation is not None:
                    product.activation = default_activation
                if correct_precursor_mz:
                    product.precursor_information.correct_mz()
            if discard_peaks:
                bunch.precursor.peak_set = None
            writer.save_scan_bunch(bunch)
    writer.complete()


@ms_conversion.command("mzml", short_help="Convert a mass spectrometry data file to mzML")
@click.argument("source")
@click.argument("output", type=click.Path(writable=True))
@click.option("-r", "--ms1-filter", "ms1_filters", multiple=True, type=parse_filter)
@click.option("-rn", "--msn-filter", "msn_filters", multiple=True, type=parse_filter)
@click.option("-p", "--pick-peaks", is_flag=True, help=("Enable peak picking, centroiding profile data"))
@click.option("-f", "--reprofile", is_flag=True, help=(
    "Enable reprofiling, converting all MS1 spectra into smoothed profile data"))
@click.option("-z", "--compress", is_flag=True, help=("Compress the output file using gzip"))
@click.option("-c", "--correct-precursor-mz", is_flag=True, help=(
    "Adjust the precursor m/z of each MSn scan to the nearest peak m/z in the precursor"))
@click.option("--update-metadata/--no-update-metadata", default=True, help=(
    "Whether or not to add the conversion"
    " program's metadata to the mzML file."))
@click.option('-X', '--option', 'options', type=(str, type_cast), multiple=True,
              help="Pass an arbitrary key-value pair through to configure the writer")
def mzml(source, output, ms1_filters=None, msn_filters=None, pick_peaks=False, reprofile=False, compress=False,
         correct_precursor_mz=False, update_metadata=True, options=None):
    """Convert `source` into mzML format written to `output`, applying a collection of optional data
    transformations along the way.
    """
    use_index = True
    if source == "-":
        click.secho("Reading input file from STDIN, some file formats will not be supported.", err=True, fg='yellow')
        source = PreBufferedStreamReader(click.open_file(source, mode='rb'))
        # Cannot use the offset index of a file we cannot seek through
        use_index = False

    reader = ms_deisotope.MSFileLoader(source, use_index=use_index)
    is_a_tty = False
    if compress:
        if not output.endswith(".gz") and output != '-':
            output += '.gz'
        stream = click.open_file(output, 'wb')
        stream = GzipFile(fileobj=stream, mode='wb')
    else:
        stream = click.open_file(output, 'wb')

    opt_map = {}
    for name, value in options:
        opt_map[name] = value
    try:
        is_a_tty = stream.isatty()
    except AttributeError: # Not all file-like objects have this method...
        if output == "-":
            is_a_tty = True
    if is_a_tty:
        write_index = False
    else:
        write_index = True
    with stream:
        to_mzml(reader, stream, pick_peaks=pick_peaks, reprofile=reprofile, ms1_filters=ms1_filters,
                msn_filters=msn_filters, correct_precursor_mz=correct_precursor_mz,
                write_index=write_index, update_metadata=update_metadata, close=False, **opt_map)


try:
    from psims.mzmlb.writer import DEFAULT_COMPRESSOR
    @ms_conversion.command("mzmlb", short_help="Convert a mass spectrometry data file to mzMLb")
    @click.argument("source")
    @click.argument("output", type=click.Path(writable=True), required=False)
    @click.option("-r", "--ms1-filter", "ms1_filters", multiple=True, type=parse_filter)
    @click.option("-rn", "--msn-filter", "msn_filters", multiple=True, type=parse_filter)
    @click.option("-p", "--pick-peaks", is_flag=True, help=("Enable peak picking, centroiding profile data"))
    @click.option("-f", "--reprofile", is_flag=True, help=(
        "Enable reprofiling, converting all MS1 spectra into smoothed profile data"))
    @click.option("-c", "--correct-precursor-mz", is_flag=True, help=(
        "Adjust the precursor m/z of each MSn scan to the nearest peak m/z in the precursor"))
    @click.option("--update-metadata/--no-update-metadata", default=True, help=(
        "Whether or not to add the conversion"
        " program's metadata to the mzML file."))
    @click.option("-z", "--compression", type=click.Choice(
        ['gzip', 'blosc', 'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd', 'zlib']), default=DEFAULT_COMPRESSOR,
        help="The compressor to use")
    @click.option('-X', '--option', 'options', type=(str, type_cast), multiple=True,
              help="Pass an arbitrary key-value pair through to configure the writer")
    def mzmlb(source, output, ms1_filters=None, msn_filters=None, pick_peaks=False, reprofile=False,
              correct_precursor_mz=False, update_metadata=True, compression=DEFAULT_COMPRESSOR, options=None):
        """Convert `source` into mzML format written to `output`, applying a collection of optional data
        transformations along the way.
        """
        use_index = True
        if source == "-":
            click.secho(
                "Reading input file from STDIN, some file formats will not be supported.",
                err=True,
                fg='yellow'
            )
            source = PreBufferedStreamReader(click.open_file(source, mode='rb'))
            # Cannot use the offset index of a file we cannot seek through
            use_index = False

        reader = ms_deisotope.MSFileLoader(source, use_index=use_index)
        if output == '-':
            raise ValueError("Cannot write HDF5 to STDOUT")
        elif output is None:
            if reader.source_file_name.endswith('.gz'):
                out = reader.source_file_name[:-3]
            else:
                out = reader.source_file_name
            out = os.path.basename(out)
            base, _ext = out.rsplit(".", 1)
            base += '.mzMLb'
            output = base

        opt_map = {}
        for name, value in options:
            opt_map[name] = value
        write_index = False
        to_mzml(reader, output, pick_peaks=pick_peaks, reprofile=reprofile, ms1_filters=ms1_filters,
                msn_filters=msn_filters, correct_precursor_mz=correct_precursor_mz,
                write_index=write_index, update_metadata=update_metadata,
                writer_type=MzMLbSerializer, compression=compression, **opt_map)
except ImportError:
    pass


if is_debug_mode():
    register_debug_hook()

if __name__ == '__main__':
    click.secho("Running Debug Mode", fg='yellow')
    register_debug_hook()
    ms_conversion.main()
